# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Collection of predefined transformations for the :py:class:`~masci_tools.io.parsers.hdf5.reader.HDF5Reader` class

All Transformation have to be able to handle (or fail gracefully with a clear error)
for the following 3 cases:

    1. The dataset is still a h5py.Dataset and might need to be transformed to a numpy array
    2. The dataset is a numpy array
    3. The dataset is a dict. This is needed to read arbitrary child dataset, where not all labels
       are known. Two options can be chosen apply the transformation to all keys in the dict
       or throw an error
"""
import h5py
import numpy as np
from functools import wraps
from .reader import HDF5Reader


class HDF5TransformationError(Exception):
    pass


def hdf5_transformation(*, attribute_needed):
    """
    Decorator for registering a function as a transformation functions
    on the :py:class:`~masci_tools.io.parsers.hdf5.reader.HDF5Reader` class

    :param attribute_needed: bool if True this function takes a previously processed
                             attribute value and is therefore only available for the entries in datasets
    """

    def hdf5_transformation_decorator(func):
        """
        Return decorated HDF5Reader object with _transforms dict and
        _attribute_transforms set attribute
        Here all registered transforms are inserted
        """

        @wraps(func)
        def transform_func(*args, **kwargs):
            """Decorator for transformation function"""
            try:
                return func(*args, **kwargs)
            except Exception as exc:
                raise HDF5TransformationError(f'The HDF5 transformation {func.__name__} failed with {exc}') from exc

        if getattr(HDF5Reader, '_transforms', None) is None:
            HDF5Reader._transforms = {}  # pylint: disable=protected-access
            HDF5Reader._attribute_transforms = set()  # pylint: disable=protected-access

        HDF5Reader._transforms[func.__name__] = transform_func  # pylint: disable=protected-access

        if attribute_needed:
            HDF5Reader._attribute_transforms.add(func.__name__)  # pylint: disable=protected-access

        return transform_func

    return hdf5_transformation_decorator


@hdf5_transformation(attribute_needed=False)
def get_first_element(dataset):
    """
    Get the first element of the dataset.

    :param dataset: dataset to transform

    :returns: first element of the dataset
    """
    return index_dataset(dataset, 0)


@hdf5_transformation(attribute_needed=False)
def index_dataset(dataset, index):
    """
    Get the n-th element of the dataset.

    :param dataset: dataset to transform

    :returns: first element of the dataset
    """
    if isinstance(dataset, dict):
        transformed = {key: data[index] for key, data in dataset.items()}
    else:
        transformed = dataset[index]

    return transformed


@hdf5_transformation(attribute_needed=False)
def slice_dataset(dataset, slice_arg):
    """
    Slice the dataset with the given slice argument.

    :param dataset: dataset to transform
    :param slice_arg: slice to apply to the dataset

    :returns: first element of the dataset
    """
    if isinstance(dataset, dict):
        transformed = {key: data[slice_arg] for key, data in dataset.items()}
    else:
        transformed = dataset[slice_arg]

    return transformed


@hdf5_transformation(attribute_needed=False)
def get_shape(dataset):
    """
    Get the shape of the dataset.

    :param dataset: dataset to get the shape

    :returns: shape of the dataset
    """

    if isinstance(dataset, dict):
        transformed = {key: data.shape for key, data in dataset.items()}
    else:
        transformed = dataset.shape

    return transformed


@hdf5_transformation(attribute_needed=False)
def get_name(dataset, full_path=False):
    """
    Get the name of the dataset.

    :param dataset: dataset to get the shape
    :param full_path: bool, if True the full path to the dataset is returned

    :returns: name of the dataset
    """

    if isinstance(dataset, (list, np.ndarray)):
        raise ValueError('Dataset has to be a h5py.Dataset for get_name')

    if isinstance(dataset, dict):
        if full_path:
            transformed = {key: data.name for key, data in dataset.items()}
        else:
            transformed = {key: data.name.split('/')[-1] for key, data in dataset.items()}
    elif full_path:
        transformed = dataset.name
    else:
        transformed = dataset.name.split('/')[-1]

    return transformed


@hdf5_transformation(attribute_needed=False)
def tile_array(dataset, n_repeats):
    """
    Use numpy.tile to repeat array n-times

    :param dataset: dataset to transform
    :param attribute_shape: int, time sto repeat the given array

    :returns: dataset repeated n-times
    """
    if isinstance(dataset, dict):
        transformed = {key: np.tile(dataset, n_repeats) for key, data in dataset.items()}
    else:
        transformed = np.tile(dataset, n_repeats)

    return transformed


@hdf5_transformation(attribute_needed=False)
def repeat_array(dataset, n_repeats):
    """
    Use numpy.repeat to repeat each element in array n-times

    :param dataset: dataset to transform
    :param n_repeats: int, time to repeat each element

    :returns: dataset with elements repeated n-times
    """
    if isinstance(dataset, dict):
        transformed = {key: np.repeat(dataset, n_repeats) for key, data in dataset.items()}
    else:
        transformed = np.repeat(dataset, n_repeats)

    return transformed


@hdf5_transformation(attribute_needed=False)
def get_all_child_datasets(group, ignore=None, contains=None):
    """
    Get all datasets contained in the given group

    :param group: h5py object to extract from
    :param ignore: str or iterable of str (optional). These
                   keys will be ignored

    :returns: a dict with the contained dataset entered with their names as keys
    """
    if ignore is None:
        ignore = set()

    if isinstance(ignore, str):
        ignore = set([ignore])

    transformed = {}
    for key, val in group.items():
        if key in ignore:
            continue
        if contains is not None:
            if contains not in key:
                continue
        if isinstance(val, h5py.Dataset):
            transformed[key] = val

    return transformed


@hdf5_transformation(attribute_needed=False)
def shift_dataset(dataset, scalar_value, negative=False):
    """
    Shift the dataset by the given scalar_value

    :param dataset: dataset to transform
    :param scalar_value: value to shift the dataset by
    :param negative: bool, if True the scalar_value will be substracted

    :returns: the dataset shifted by the scalar
              if it is a dict all entries are shifted
    """
    transformed = dataset
    if isinstance(transformed, dict):
        transformed = {
            key: np.array(entry) if isinstance(entry, h5py.Dataset) else entry for key, entry in transformed.items()
        }
    elif isinstance(transformed, h5py.Dataset):
        transformed = np.array(transformed)

    if isinstance(transformed, dict):
        if negative:
            transformed = {key: entry - scalar_value for key, entry in transformed.items()}
        else:
            transformed = {key: entry + scalar_value for key, entry in transformed.items()}
    else:
        if negative:
            transformed = transformed - scalar_value
        else:
            transformed = transformed + scalar_value

    return transformed


@hdf5_transformation(attribute_needed=False)
def multiply_scalar(dataset, scalar_value):
    """
    Multiply the given dataset with a scalar_value

    :param dataset: dataset to transform
    :param scalar_value: value to mutiply the dataset by

    :returns: the dataset multiplied by the scalar
              if it is a dict all entries are multiplied
    """
    transformed = dataset
    if isinstance(transformed, dict):
        transformed = {
            key: np.array(entry) if isinstance(entry, h5py.Dataset) else entry for key, entry in transformed.items()
        }
    elif isinstance(transformed, h5py.Dataset):
        transformed = np.array(transformed)

    if isinstance(transformed, dict):
        transformed = {key: entry * scalar_value for key, entry in transformed.items()}
    else:
        transformed = transformed * scalar_value

    return transformed


@hdf5_transformation(attribute_needed=False)
def convert_to_complex_array(dataset):
    """
    Converts the given dataset of real numbers into
    dataset of complex numbers. This follows the convention of
    how complex numbers are normally written out by Fleur
    (last index 0 real part, last index 1 imag part)

    :param dataset: dataset to transform

    :returns: dataset with complex values
    """
    transformed = dataset
    if isinstance(transformed, dict):
        transformed = {
            key: np.array(entry) if isinstance(entry, h5py.Dataset) else entry for key, entry in transformed.items()
        }
    elif isinstance(transformed, h5py.Dataset):
        transformed = np.array(transformed)

    if isinstance(transformed, dict):
        transformed = {key: entry[..., 0] + 1j * entry[..., 1] for key, entry in transformed.items()}
    else:
        transformed = transformed[..., 0] + 1j * transformed[..., 1]

    return transformed


@hdf5_transformation(attribute_needed=False)
def multiply_array(dataset, matrix, transpose=False):
    """
    Multiply the given dataset with a matrix

    :param dataset: dataset to multiply
    :param matrix: matrix to multiply by
    :param transpose: bool, if True the given matrix is transposed

    :returns: dataset multiplied with the given matrix
    """

    transformed = dataset

    if isinstance(transformed, dict):
        transformed = {
            key: np.array(entry) if isinstance(entry, h5py.Dataset) else entry for key, entry in transformed.items()
        }
    elif isinstance(transformed, h5py.Dataset):
        transformed = np.array(dataset)

    if transpose:
        matrix = matrix.T

    if isinstance(transformed, dict):
        transformed = {key: entry.dot(matrix) for key, entry in transformed.items()}
    else:
        transformed = transformed.dot(matrix)

    return transformed


@hdf5_transformation(attribute_needed=False)
def calculate_norm(dataset, between_neighbours=False):
    """
    Calculate norms on the given dataset. Calculates the norm of each row in the dataset

    :param dataset: dataset to transform
    :param between_neighbours: bool, if True the distance between subsequent entries in the dataset is calculated

    :returns: norms of the given dataset
    """
    transformed = dataset
    if isinstance(dataset, h5py.Dataset):
        transformed = np.array(dataset)

    if isinstance(dataset, dict):
        raise NotImplementedError

    if between_neighbours:
        transformed = np.array([np.linalg.norm(ki - kj) for ki, kj in zip(transformed[1:], transformed[:-1])])
    else:
        transformed = np.linalg.norm(transformed, axis=0)

    return transformed


@hdf5_transformation(attribute_needed=False)
def cumulative_sum(dataset, beginning_zero=True):
    """
    Calculate the cumulative sum of the dataset

    :param dataset: dataset to transform

    :returns: cumulative sum of the dataset
    """

    if isinstance(dataset, dict):
        raise NotImplementedError

    transformed = dataset
    if isinstance(dataset, h5py.Dataset):
        transformed = np.array(dataset)

    if beginning_zero:
        transformed = np.insert(transformed, 0, 0.0)

    transformed = np.cumsum(transformed)

    return transformed


@hdf5_transformation(attribute_needed=False)
def get_attribute(dataset, attribute_name):
    """Extracts a specified attribute's value.

    :param dataset: dataset to transform
    :param attribute_name: str of the attribute to extract from the dataset

    :returns: value of the attribute on the dataset
    """
    if isinstance(dataset, (np.ndarray, dict)):
        raise NotImplementedError

    transformed = dataset.attrs[attribute_name]

    return transformed


@hdf5_transformation(attribute_needed=False)
def attributes(dataset):
    """Extracts all attributes of the dataset

    :param dataset: dataset to transform

    :returns: dict with all the set attributes on the dataset
    """
    if isinstance(dataset, (np.ndarray, dict)):
        raise NotImplementedError

    transformed = dict(dataset.attrs)

    return transformed


@hdf5_transformation(attribute_needed=False)
def move_to_memory(dataset):
    """Moves the given dataset to memory, if it's not already there
    Creates numpy arrays for each dataset it finds

    :param dataset: dataset to transform

    :returns: dataset with h5py.Datasets converted to numpy arrays
    """

    transformed = dataset
    if isinstance(transformed, dict):
        for key, val in transformed.items():
            transformed[key] = move_to_memory(val)
    elif isinstance(transformed, h5py.Dataset):
        transformed = np.array(dataset)

    return transformed


@hdf5_transformation(attribute_needed=False)
def flatten_array(dataset, order='C'):
    """
    Flattens the given dataset to one dimensional array.
    Copies the array !!

    :param dataset: dataset to transform
    :param order: str {‘C’, ‘F’, ‘A’, ‘K’} flatten in column major
                  or row-major order (see numpy.flatten documentation)

    :returns: flattened dataset
    """

    transformed = dataset
    if isinstance(transformed, dict):
        transformed = {
            key: np.array(entry) if isinstance(entry, h5py.Dataset) else entry for key, entry in transformed.items()
        }
    elif isinstance(transformed, h5py.Dataset):
        transformed = np.array(transformed)

    if isinstance(transformed, dict):
        transformed = {key: entry.flatten(order=order) for key, entry in transformed.items()}
    else:
        transformed = transformed.flatten(order=order)

    return transformed


@hdf5_transformation(attribute_needed=False)
def split_array(dataset, suffixes=None, name=None):
    """
    Split the arrays in a dataset into multiple entries
    by their first index

    If the dataset is a dict the entries will be split up.
    If the dataset is not a dict a dict is created with the dataset
    entered under `name` and this will be split up

    :param dataset: dataset to transform
    :param suffix: Optional list of str to use for suffixes
                   for the split up entries. by default it is
                   the value of the first index of the original
                   array
    :param name: str for the case of the dataset not being a
                 dict. Key for the entry in the new dict for
                 the original dataset. The returned dataset will only
                 contain the split up entries

    :param dataset: dict with the entries split up
    """

    if not isinstance(dataset, dict) and name is None:
        raise ValueError('split_arrays has to be given a name if the dataset is not a dict')

    transformed = dataset
    if not isinstance(dataset, dict):
        transformed = {name: dataset}

    transformed = {
        key: np.array(entry) if isinstance(entry, h5py.Dataset) else entry for key, entry in transformed.items()
    }
    max_length = max(len(entry) for entry in transformed.values())

    if suffixes is None:
        suffixes = list(range(max_length))

    if len(suffixes) < max_length:
        raise ValueError(f'Too few suffixes provided: Expected {max_length} Got: {len(suffixes)}')

    for key in list(transformed.keys()):
        val = transformed.pop(key)
        for suffix, entry in zip(suffixes, val):
            transformed[f'{key}_{suffix}'] = entry

    return transformed


@hdf5_transformation(attribute_needed=False)
def convert_to_str(dataset):
    """Converts the given dataset to a numpy array of type string

    :param dataset: dataset to transform

    :returns: numpy array of dtype str
    """

    transformed = np.array(dataset).astype(str)

    return transformed


@hdf5_transformation(attribute_needed=False)
def apply_lambda(dataset, lambda_func):
    """Applies a given lambda function to the dataset
       This should be used with care. One possible example is
       converting to a boolean with lambda x: x==1

    :param dataset: dataset to transform
    :param lambda_func: lambda function to apply to the dataset

    :returns: return value of the lambda function
    """

    transformed = dataset
    if isinstance(transformed, dict):
        transformed = {key: lambda_func(entry) for key, entry in transformed.items()}
    else:
        transformed = lambda_func(dataset)

    return transformed


@hdf5_transformation(attribute_needed=False)
def periodic_elements(dataset):
    """Converts the given dataset (int or list of ints)
       To the atomic symbols corresponding to the atomic number

    :param dataset: dataset to transform

    :returns: str or array of str with the atomic elements
    """
    from masci_tools.util.constants import PERIODIC_TABLE_ELEMENTS

    if isinstance(dataset, dict):
        raise NotImplementedError

    if isinstance(dataset, int):
        transformed = PERIODIC_TABLE_ELEMENTS[dataset]['symbol']
    else:
        transformed = np.array([PERIODIC_TABLE_ELEMENTS[entry]['symbol'] for entry in dataset], dtype=str)

    return transformed


@hdf5_transformation(attribute_needed=False)
def sum_over_dict_entries(dataset, overwrite_dict=False, entries=None, dict_entry='sum', entry_format=None):
    """
    Sum the datasets contained in the given dict dataset

    :param dataset: dataset to transform
    :param overwrite_dict: bool if True, the result will overwrite the dictionary
                           if False it is entered under `sum` in the dict

    :returns: dataset with summed entries
    """

    if not isinstance(dataset, dict):
        raise ValueError('sum_over_dict_entries is only available for dict datasets')

    if entries is None:
        entries = dataset.keys()

    if entry_format is not None:
        entries = [entry_format(entry) for entry in entries]

    transformed = dataset
    if overwrite_dict:
        transformed = np.sum([entry for key, entry in dataset.items() if key in entries], axis=0)
    else:
        transformed[dict_entry] = np.sum([entry for key, entry in dataset.items() if key in entries], axis=0)

    return transformed


@hdf5_transformation(attribute_needed=False)
def add_partial_sums_fixed(dataset, patterns, replace_entries=None):
    """
    Add entries to the dataset dict (Only avalaible for dict datasets) with sums
    over entries containing a given pattern

    Used for example in the FleurBands recipe to calculate total atom weights
    with the patterns `['MT:1', 'MT:2', ...]`

    :param dataset: dataset to transform
    :param patterns: list of str to sum entries over
    :param replace_entries: list of str under which to enter the entries back

    :returns: dataset with new entries containing the sums over entries matching the given pattern
    """
    from collections.abc import Iterable

    if not isinstance(dataset, dict):
        raise ValueError('add_partial_sums_fixed only available for dict datasets')

    if not isinstance(patterns, Iterable) and not isinstance(patterns, str):
        raise ValueError('patterns has be an Iterable')

    if replace_entries is None:
        replace_entries = patterns

    transformed = dataset.copy()
    for pattern, replace_entry in zip(patterns, replace_entries):
        entries = [key for key in transformed.keys() if pattern in key]

        transformed[replace_entry] = sum_over_dict_entries(dataset, overwrite_dict=True, entries=entries)

    return transformed


#Functions that can use an attribute value (These are passed in from _transform_dataset)
#The transformation don't have access to all the attributes


@hdf5_transformation(attribute_needed=True)
def multiply_by_attribute(dataset, attribute_value, transpose=False):
    """
    Multiply the given dataset with a previously parsed attribute, either scalar or matrix like

    :param dataset: dataset to transform
    :param attribute_value: value to multiply by (attribute value passed in from `_transform_dataset`)

    Only relevant for matrix multiplication:
        :param transpose: bool if True the Matrix order is transposed before multiplying

    :returns: dataset multiplied with the given attribute_value
    """
    if isinstance(attribute_value, h5py.Dataset):
        attribute_value = np.array(attribute_value)

    if isinstance(attribute_value, np.ndarray):
        transformed = multiply_array(dataset, attribute_value, transpose=transpose)
    else:
        transformed = multiply_scalar(dataset, attribute_value)

    return transformed


@hdf5_transformation(attribute_needed=True)
def shift_by_attribute(dataset, attribute_value, negative=False):
    """
    Shift the dataset by the given value of the attribute

    :param dataset: dataset to transform
    :param attribute_value: value to shift the dataset by
    :param negative: bool, if True the scalar_value will be substracted

    :returns: the dataset shifted by the scalar
              if it is a dict all entries are shifted
    """
    return shift_dataset(dataset, attribute_value, negative=negative)


@hdf5_transformation(attribute_needed=True)
def add_partial_sums(dataset, attribute_value, pattern_format, make_set=False, replace_entries=None):
    """
    Add entries to the dataset dict (Only avalaible for dict datasets) with sums
    over entries containing a given pattern formatted with a attribute_value

    Used for example in the FleurBands recipe to calculate total atom weights
    with the pattern_format `'MT:{}'.format` and the atomtype as the attribute_value

    :param dataset: dataset to transform
    :param attribute_value: value to multiply by (attribute value passed in from `_transform_dataset`)
    :param pattern_format: callable returning a formatted string
                           This will be called with every entry in the attribute_value list
    :param replace_entries: list of str under which to enter the entries back

    :returns: dataset with new entries containing the sums over entries matching the given pattern
    """
    if isinstance(attribute_value, h5py.Dataset):
        attribute_value = np.array(attribute_value)

    if not isinstance(dataset, dict):
        raise ValueError('add_partial_sums only available for dict datasets')

    if not isinstance(attribute_value, (list, np.ndarray)):
        raise ValueError('attribute_value has be a list or array')

    if make_set:
        attribute_value = set(attribute_value)

    return add_partial_sums_fixed(dataset, [pattern_format(val) for val in attribute_value],
                                  replace_entries=replace_entries)


@hdf5_transformation(attribute_needed=True)
def repeat_array_by_attribute(dataset, attribute_value):
    """
    Use numpy.repeat to repeat each element in array n-times (given by attribute_value)

    :param dataset: dataset to transform
    :param attribute_shape: int, time to repeat the elements in the given array

    :returns: dataset with elements repeated n-times
    """
    return repeat_array(dataset, attribute_value)


@hdf5_transformation(attribute_needed=True)
def tile_array_by_attribute(dataset, attribute_value):
    """
    Use numpy.tile to repeat array n-times (given by attribute_value)

    :param dataset: dataset to transform
    :param attribute_shape: int, time sto repeat the given array

    :returns: dataset repeated n-times
    """
    return tile_array(dataset, attribute_value)



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Here commonly used functions that do not need aiida-stuff (i.e. can be tested
without a database) are collected.
"""
import io
import numpy as np
import warnings
from collections import namedtuple
####################################################################################

#helper functions used in calculation, parser etc.


def open_general(filename_or_handle, iomode=None):
    """
    Open a file directly from a path or use a file handle if that is given.
    Also take care of closed files by reopenning them.
    This is intended to be used like this::

        f = open_general(outfile)
        with f: # make sure the file is properly closed
            txt = f.readlines()
    """
    reopen_file = False
    # this is needed in order to make python2 and 3 work (in py3 file does not exist anymore)

    if not isinstance(filename_or_handle, io.IOBase):
        reopen_file = True

    if reopen_file:
        if iomode is None:
            iomode = u'r'
        f = open(filename_or_handle, iomode)
    else:
        f = filename_or_handle
        if f.closed:  # reopen file if it was closed before
            if iomode is None:
                f = open(f.name, f.mode)
            else:
                f = open(f.name, iomode)
        else:  # make sure reading the file now starts at the beginning again
            f.seek(0)
    return f


def skipHeader(seq, n):
    """Iterate over a sequence skipping the first n elements

    Args:
        seq (iterable): Iterable sequence
        n (int): Number of Elements to skip in the beginning of the sequence

    Yields:
        item: Elements in seq after the first n elements
    """
    for i, item in enumerate(seq):
        if i >= n:
            yield item


def filter_out_empty_dict_entries(dict_to_filter):
    """
    Filter out entries in a given dict that correspond to empty values.
    At the moment this is empty lists, dicts and None

    :param dict_to_filter: dict to filter

    :returns: dict without empty entries
    """

    EMPTY_VALUES = (None, [], {})

    return {key: val for key, val in dict_to_filter.items() if val not in EMPTY_VALUES}


def get_alat_from_bravais(bravais, is3D=True):
    bravais_tmp = bravais
    if not is3D:
        #take only in-plane lattice to find maximum as alat
        bravais_tmp = bravais[:2, :2]
    return np.sqrt(np.sum(bravais_tmp**2, axis=1)).max()


def search_string(searchkey, txt):
    iline = 0
    for line in txt:
        if searchkey in line:
            return iline
        iline += 1
    return -1


def angles_to_vec(magnitude, theta, phi):
    """
    convert (magnitude, theta, phi) to (x,y,z)

    theta/phi need to be in radians!

    Input can be single number, list of numpy.ndarray data
    Returns x,y,z vector
    """

    # correct data type if necessary
    if isinstance(magnitude, list):
        magnitude = np.array(magnitude)
    if isinstance(theta, list):
        theta = np.array(theta)
    if isinstance(phi, list):
        phi = np.array(phi)
    single_value_input = False
    if not isinstance(magnitude, np.ndarray):
        magnitude = np.array([magnitude])
        single_value_input = True
    if not isinstance(theta, np.ndarray):
        theta = np.array([theta])
        single_value_input = True
    if not isinstance(phi, np.ndarray):
        phi = np.array([phi])
        single_value_input = True

    vec = []
    for mag_i, phi_i, theta_i in zip(magnitude, phi, theta):
        r_inplane = mag_i * np.sin(theta_i)
        x = r_inplane * np.cos(phi_i)
        y = r_inplane * np.sin(phi_i)
        z = np.cos(theta_i) * mag_i
        vec.append([x, y, z])
    vec = np.array(vec)

    if single_value_input:
        vec = vec[0]

    return vec


def get_Ang2aBohr():
    from masci_tools.util.constants import ANG_BOHR_KKR
    #warnings.warn(
    #    'get_Ang2aBohr is deprecated. Use 1/BOHR_A with the BOHR_A constant from the module masci_tools.util.constants instead',
    #    DeprecationWarning)
    return ANG_BOHR_KKR


def get_aBohr2Ang():
    from masci_tools.util.constants import ANG_BOHR_KKR
    #warnings.warn(
    #    'get_aBohr2Ang is deprecated. Use the BOHR_A constant from the module masci_tools.util.constants instead',
    #    DeprecationWarning)
    return 1.0 / ANG_BOHR_KKR


def get_Ry2eV():
    from masci_tools.util.constants import RY_TO_EV_KKR
    #warnings.warn(
    #    'get_Ry2eV is deprecated. Use the RY_TO_EV constant from the module masci_tools.util.constants instead',
    #    DeprecationWarning)
    return RY_TO_EV_KKR


def vec_to_angles(vec):
    """
    converts vector (x,y,z) to (magnitude, theta, phi)
    """
    magnitude, theta, phi = [], [], []
    if len(vec) == 3 and len(np.shape(vec)) < 2:
        vec = np.array([vec])
        multiple_entries = False
    else:
        multiple_entries = True

    for vec_i in vec:
        phi.append(np.arctan2(vec_i[1], vec_i[0]))
        r_inplane = np.sqrt(vec_i[0]**2 + vec_i[1]**2)
        theta.append(np.arctan2(r_inplane, vec_i[2]))
        magnitude.append(np.sqrt(r_inplane**2 + vec_i[2]**2))
    if multiple_entries:
        magnitude, theta, phi = np.array(magnitude), np.array(theta), np.array(phi)
    else:
        magnitude, theta, phi = magnitude[0], theta[0], phi[0]
    return magnitude, theta, phi


def get_version_info(outfile):
    f = open_general(outfile)
    with f:
        tmptxt = f.readlines()
    itmp = search_string('Code version:', tmptxt)
    if itmp == -1:  # try to find serial number from header of file
        itmp = search_string('# serial:', tmptxt)
        code_version = tmptxt[itmp].split(':')[1].split('_')[1].strip()
        compile_options = tmptxt[itmp].split(':')[1].split('_')[2].strip()
        serial_number = tmptxt[itmp].split(':')[1].split('_')[3].strip()
    else:
        code_version = tmptxt.pop(itmp).split(':')[1].strip()
        itmp = search_string('Compile options:', tmptxt)
        compile_options = tmptxt.pop(itmp).split(':')[1].strip()
        itmp = search_string('serial number for files:', tmptxt)
        serial_number = tmptxt.pop(itmp).split(':')[1].strip()
    return code_version, compile_options, serial_number


def get_corestates_from_potential(potfile='potential'):
    """Read core states from potential file"""
    f = open_general(potfile)
    with f:
        txt = f.readlines()

    #get start of each potential part
    istarts = [iline for iline in range(len(txt)) if 'POTENTIAL' in txt[iline]]
    print(istarts)

    n_core_states = []  #number of core states per potential
    e_core_states = []  #energies of core states
    l_core_states = []  #angular momentum index, i.e. 0=s, 1=p etc...
    for pot_index, start_index in enumerate(istarts):
        line = txt[start_index + 6]
        n = int(line.split()[0])
        print(pot_index, n)
        n_core_states.append(n)
        elevels = np.zeros(n)  #temp array for energies
        langmom = np.zeros(n, dtype=int)  #temp array for angular momentum index
        for icore in range(n):
            line = txt[start_index + 7 + icore].split()
            langmom[icore] = int(line[0])
            elevels[icore] = float(line[1].replace('D', 'E'))
        e_core_states.append(elevels)
        l_core_states.append(langmom)

    return n_core_states, e_core_states, l_core_states


def get_highest_core_state(nstates, energies, lmoments):
    """Find highest lying core state from list of core states, needed to find and check energy contour"""
    idx = energies.argmax()
    lval = lmoments[idx]
    nquant = sum(lmoments == lval) + lval
    level_descr = '%i%s' % (nquant, 'spdfgh'[lval])

    return lval, energies[idx], level_descr


def interpolate_dos(
    dosfile,
    return_original=False,
):
    r"""
    interpolation function copied from complexdos3 fortran code

    Principle of DOS here: Two-point contour integration
    for DOS in the middle of the two points. The input DOS
    and energy must be complex. Parameter deltae should be
    of the order of magnitude of eim::

              <-2*deltae->   _
                   /\        |     DOS=(n(1)+n(2))/2 + (n(1)-n(2))*eim/deltae
                  /  \       |
                (1)  (2)   2*i*eim=2*i*pi*Kb*Tk
                /      \     |
               /        \    |
        ------------------------ (Real E axis)

    :param input: either absolute path of 'complex.dos' file or file handle to it

    :returns: E_Fermi, numpy array of interpolated dos

    :note: output units are in Ry!
    """

    with open_general(dosfile) as f:
        text = f.readline()  # dummy readin of header, may be replaced later
        npot = int(f.readline().split()[0])
        iemax = int(f.readline().split()[0])
        lmax = int(f.readline().split()[0])

        dosnew_all_atoms = []
        dos_all_atoms = []

        for i1 in range(npot):
            #print('Reading potential',i1)
            # Read header (not used)
            for iheader in range(3):
                text = f.readline()

            # extract EF
            ef = float(f.readline().split()[7])

            # some more dummy lines
            for iheader in range(5, 9 + 1):
                text = f.readline()

            # now header is done. start reading DOS
            # Read dos: (total dos stored at DOS(LMAX+1,IE))
            dos_l_cmplx = []
            for ie in range(iemax):
                tmpline = f.readline().replace('(', '').replace(')', '').replace(',', '').split()
                ez = float(tmpline[0]) + 1j * float(tmpline[1])
                dostmp_complex = [[tmpline[len(tmpline) - 2], tmpline[len(tmpline) - 1]]]
                dostmp_complex += [[tmpline[iline], tmpline[iline + 1]] for iline in range(2, len(tmpline) - 2, 2)]
                dostmp = [ez] + [float(ds[0]) + 1j * float(ds[1]) for ds in dostmp_complex]
                dos_l_cmplx.append(dostmp)
            dos_l_cmplx = np.array(dos_l_cmplx)
            dos_l = np.imag(dos_l_cmplx.copy())
            dos_l[:, 0] = np.real(dos_l_cmplx.copy()[:, 0])
            dos_all_atoms.append(dos_l)

            # Compute and write out corrected dos at new (middle) energy points:
            dosnew = []
            ez = dos_l_cmplx[:, 0]
            for ie in range(1, iemax - 1):
                deltae = np.real(ez[ie + 1] - ez[ie])
                eim = np.imag(ez[ie])
                enew = np.real(ez[ie])  # Real quantity

                tmpdos = [enew]
                for ll in range(1, lmax + 3):
                    t = (dos_l_cmplx[ie - 1, ll] - dos_l_cmplx[ie + 1, ll]) * 0.5 * (0.0 + eim * 1j) / deltae
                    #print ie+1, ll,  dos_l_cmplx[ie, ll], deltae, eim, t, shape(dos_l_cmplx[ie]), lmax
                    #tmpdos.append(dos_l_cmplx[ie, ll] + 0.5*(dos_l_cmplx[ie-1, ll]-dos_l_cmplx[ie+1, ll])*(0.+1j*eim)/deltae)
                    tmpdos.append(dos_l_cmplx[ie, ll] + t)
                tmpdos = np.array(tmpdos)
                # build imaginary part (factor -1/2pi is already included)
                tmpdos = np.array([np.real(tmpdos[0])] + [np.imag(ds) for ds in tmpdos[1:]])
                dosnew.append(tmpdos)

            # save to big array with all atoms
            dosnew_all_atoms.append(dosnew)

            if i1 != npot:
                text = f.readline()  # dummy line

        dosnew_all_atoms = np.array(dosnew_all_atoms)
        dos_all_atoms = np.array(dos_all_atoms)

    if return_original:
        return ef, dos_all_atoms, dosnew_all_atoms
    else:
        return ef, dosnew_all_atoms


def get_ef_from_potfile(potfile):
    """
    extract fermi energy from potfile
    """
    with open_general(potfile) as f:
        txt = f.readlines()
    ef = float(txt[3].split()[1])
    return ef


def convert_to_pystd(value):
    """Recursively convert numpy datatypes to standard python, needed by aiida-core.

    Usage:
        converted = convert_to_pystd(to_convert)

    where `to_convert` can be a dict, array, list, or single valued variable
    """
    if isinstance(value, np.ndarray):
        value = list(value)
        value = convert_to_pystd(value)
    elif isinstance(value, list):
        for index, val in enumerate(value):
            value[index] = convert_to_pystd(val)
    elif isinstance(value, tuple):
        value = tuple(convert_to_pystd(val) for val in value)
    elif isinstance(value, np.integer):
        value = int(value)
    elif isinstance(value, np.floating):
        value = float(value)
    elif isinstance(value, np.str_):
        value = str(value)
    elif isinstance(value, dict):
        for key, val in value.items():
            value[key] = convert_to_pystd(val)
    return value


def camel_to_snake(name):
    """
    Converts camelCase to snake_case variable names
    Used in the Fleur parser to convert attribute names from the xml files
    """
    name = name.replace('-', '')
    return ''.join(['_' + c.lower() if c.isupper() else c for c in name]).lstrip('_')


def convert_to_fortran(val, quote_strings=True):
    """
    :param val: the value to be read and converted to a Fortran-friendly string.
    """
    # Note that bool should come before integer, because a boolean matches also
    # isinstance(...,int)
    import numbers

    if isinstance(val, (bool, np.bool_)):
        if val:
            val_str = '.true.'
        else:
            val_str = '.false.'
    elif isinstance(val, numbers.Integral):
        val_str = f'{val:d}'
    elif isinstance(val, numbers.Real):
        val_str = f'{val:18.10e}'.replace('e', 'd')
    elif isinstance(val, str):
        if quote_strings:
            val_str = f"'{val!s}'"
        else:
            val_str = f'{val!s}'
    else:
        raise ValueError(f"Invalid value '{val}' of type '{type(val)}' passed, accepts only booleans, ints, "
                         'floats and strings')

    return val_str


def convert_to_fortran_string(string):
    """
    converts some parameter strings to the format for the inpgen
    :param string: some string
    :returns: string in right format (extra "" if not already present)
    """
    if not string.strip().startswith("\"") or \
       not string.strip().endswith("\""):
        return f'"{string}"'
    else:
        return string


def is_sequence(arg):
    """
    Checks if arg is a sequence
    """
    if isinstance(arg, str):
        return False
    elif hasattr(arg, '__iter__'):
        return True
    elif not hasattr(arg, 'strip') and hasattr(arg, '__getitem__'):
        return True
    else:
        return False


def abs_to_rel(vector, cell):
    """
    Converts a position vector in absolute coordinates to relative coordinates.

    :param vector: list or np.array of length 3, vector to be converted
    :param cell: Bravais matrix of a crystal 3x3 Array, List of list or np.array
    :return: list of length 3 of scaled vector, or False if vector was not length 3
    """

    if len(vector) == 3:
        cell_np = np.array(cell)
        inv_cell_np = np.linalg.inv(cell_np)
        postionR = np.array(vector)
        # np.matmul(inv_cell_np, postionR)#
        new_rel_post = np.matmul(postionR, inv_cell_np)
        new_rel_pos = list(new_rel_post)
        return new_rel_pos
    else:
        return False


def abs_to_rel_f(vector, cell, pbc):
    """
    Converts a position vector in absolute coordinates to relative coordinates
    for a film system.

    :param vector: list or np.array of length 3, vector to be converted
    :param cell: Bravais matrix of a crystal 3x3 Array, List of list or np.array
    :param pbc: Boundary conditions, List or Tuple of 3 Boolean
    :return: list of length 3 of scaled vector, or False if vector was not length 3
    """
    # TODO this currently only works if the z-coordinate is the one with no pbc
    # Therefore if a structure with x non pbc is given this should also work.
    # maybe write a 'tranform film to fleur_film routine'?
    if len(vector) == 3:
        if not pbc[2]:
            # leave z coordinate absolute
            # convert only x and y.
            postionR = np.array(vector)
            postionR_f = np.array(postionR[:2])
            cell_np = np.array(cell)
            cell_np = np.array(cell_np[0:2, 0:2])
            inv_cell_np = np.linalg.inv(cell_np)
            # np.matmul(inv_cell_np, postionR_f)]
            new_xy = np.matmul(postionR_f, inv_cell_np)
            new_rel_pos_f = [new_xy[0], new_xy[1], postionR[2]]
            return new_rel_pos_f
        else:
            print('FLEUR can not handle this type of film coordinate')
    else:
        return False


def rel_to_abs(vector, cell):
    """
    Converts a position vector in internal coordinates to absolute coordinates
    in Angstrom.

    :param vector: list or np.array of length 3, vector to be converted
    :param cell: Bravais matrix of a crystal 3x3 Array, List of list or np.array
    :return: list of legth 3 of scaled vector, or False if vector was not lenth 3
    """
    if len(vector) == 3:
        cell_np = np.array(cell)
        postionR = np.array(vector)
        new_abs_post = np.matmul(postionR, cell_np)
        new_abs_pos = list(new_abs_post)
        return new_abs_pos
    else:
        return False


def rel_to_abs_f(vector, cell):
    """
    Converts a position vector in internal coordinates to absolute coordinates
    in Angstrom for a film structure (2D).
    """
    # TODO this currently only works if the z-coordinate is the one with no pbc
    # Therefore if a structure with x non pbc is given this should also work.
    # maybe write a 'tranform film to fleur_film routine'?
    if len(vector) == 3:
        postionR = np.array(vector)
        postionR_f = np.array(postionR[:2])
        cell_np = np.array(cell)
        cell_np = np.array(cell_np[0:2, 0:2])
        new_xy = np.matmul(postionR_f, cell_np)
        new_abs_pos_f = [new_xy[0], new_xy[1], postionR[2]]
        return new_abs_pos_f
    else:
        return False


def find_symmetry_relation(from_pos, to_pos, rotations, shifts, cell, relative_pos=False, film=False):
    """
    Find symmetry relation between the given vectors. This functions assumes
    that a symmetry relation exists otherwise an error is raised

    :param from_pos: vector to rotate
    :param to_pos: vector to rotate to
    :param rotations: list of np.arrays with the given symmetry rotations
    :param shifts: list of np.arrays with the given shifts for the symmetry operations
    :param cell: Bravais matrix of a crystal 3x3 Array, List of list or np.array
    :param relative_pos: bool if True the given vectors are assuemd to be in internal coordinates
    :param film: bool if True the vectors are assumed to be film coordinates

    :returns: tuple of rotation and shift mapping ``from_pos`` to ``to_pos``

    :raises ValueError: If no symmetry relation is found
    """

    def lattice_shifts():
        for i in range(-2, 3):
            for j in range(-2, 3):
                for k in range(-2, 3):
                    yield np.array([i, j, k], dtype=int)

    if not relative_pos:
        if film:
            from_pos = abs_to_rel_f(from_pos, cell, (True, True, False))
            to_pos = abs_to_rel_f(to_pos, cell, (True, True, False))
        else:
            from_pos = abs_to_rel(from_pos, cell)
            to_pos = abs_to_rel(to_pos, cell)

    cell = np.array(cell)
    cell_square = np.matmul(cell.T, cell)

    for rot, shift in zip(rotations, shifts):
        rot_pos = np.matmul(rot, np.array(from_pos)) + shift
        diff = rot_pos - np.array(to_pos)
        for lat_shift in lattice_shifts():
            length = np.sqrt(np.dot(np.matmul(diff + lat_shift, cell_square), diff + lat_shift))
            if length < 1e-4:
                return rot, shift

    raise ValueError(f'No symmetry relation found between {from_pos} and {to_pos}')


#namedtuple used for input output of atom sites
AtomSiteProperties = namedtuple('AtomSiteProperties', ('position', 'symbol', 'kind'))


def get_wigner_matrix(l, phi, theta):
    """Produces the wigner rotation matrix for the density matrix

    :param l: int, orbital quantum number
    :param phi: float, angle (radian) corresponds to euler angle alpha
    :param theta: float, angle (radian) corresponds to euler angle beta
    """
    d_wigner = np.zeros((7, 7), dtype=complex)
    for m in range(-l, l + 1):
        for mp in range(-l, l + 1):
            base = np.sqrt(fac(l + m) * fac(l - m) * fac(l + mp) * fac(l - mp))
            base *= np.exp(-1j * phi * mp)

            for x in range(max(0, m - mp), min(l - mp, l + m) + 1):
                denom = fac(l - mp - x) * fac(l + m - x) * fac(x) * fac(x + mp - m)

                d_wigner[m + 3, mp + 3] += base/denom * (-1)**x * np.cos(theta/2.0)**(2*l+m-mp-2*x) \
                                          * np.sin(theta/2.0)**(2*x+mp-m)

    return d_wigner


def fac(n):
    """Returns the factorial of n"""
    if n < 2:
        return 1
    else:
        return n * fac(n - 1)



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
from matplotlib import cm


def dosplot(p0='./',
            totonly=True,
            color='',
            label='',
            marker='',
            lw=2,
            ms=5,
            ls='-',
            ls_ef=':',
            lw_ef=1,
            units='Ry',
            noefline=False,
            interpol=False,
            allatoms=False,
            onespin=False,
            atoms=[],
            lmdos=False,
            lm=[],
            nofig=False,
            scale=1.0,
            shift=0,
            normalized=False,
            xyswitch=False,
            efcolor='',
            return_data=False,
            xscale=1.,
            xshift=0.0,
            yshift=0.0,
            filled=False,
            spins=2):
    """ plotting routine for dos files """
    # import dependencies
    import numpy as np
    import matplotlib.pyplot as plt
    from os import listdir
    from os.path import isdir

    plt.ion()

    # deal with input of file handle instead of path (see plot_kkr of aiida_kkr)
    if not isinstance(p0, str):
        pathname_with_file = p0.name
        p0 = pathname_with_file.replace('/dos.atom1', '')

    # read in data
    if p0[-1] != '/':
        p0 += '/'
    #if 'rel' in units: ef = float(open(p0+'potential').readlines()[3].split()[1])
    ef = float(open(p0 + 'potential').readlines()[3].split()[1])
    first = True
    for i in np.sort(listdir(p0)):
        if (((i[:8] == 'dos.atom' and 'interpol' not in i) and not interpol) or
            (interpol and i[:17] == 'dos.interpol.atom')) or ('out_ldos.atom' in i and 'out_lmdos.atom' not in i and
                                                              not interpol) or ('out_ldos.interpol.atom' in i and
                                                                                'out_lmdos.interpol.atom' and
                                                                                interpol) and not isdir(p0 + i):
            if lmdos:
                i = i.replace('out_ldos.', 'out_lmdos.')
            if not onespin or 'spin' + str(spins) not in i:
                iatom = i.replace('dos.', '').replace('interpol.',
                                                      '').replace('atom',
                                                                  '').replace('out_l',
                                                                              '').replace('=',
                                                                                          '').replace('m',
                                                                                                      '').split('_')[0]
                if atoms == [] or int(iatom) in atoms:
                    tmp = np.loadtxt(p0 + i)
                    print(p0 + i)

                    # set units
                    if 'rel' in units:
                        tmp[:, 0] = tmp[:, 0] - ef
                    if 'eV' in units:
                        tmp[:, 0] = tmp[:, 0] * 13.6
                        tmp[:, 1:] = tmp[:, 1:] / 13.6

                    tmp[:, 0] = tmp[:, 0] * xscale + xshift
                    tmp[:, 1] = tmp[:, 1] + yshift

                    if allatoms:
                        sgn = 1
                        if 'spin2' in i:
                            sgn = -1
                        # plot dos
                        if totonly:
                            if color == '':
                                if first:
                                    if not nofig:
                                        plt.figure()
                                plt.plot(tmp[:, 0], sgn * tmp[:, 1], marker + ls, label=label + str(i), lw=lw, ms=ms)
                            else:
                                if not nofig:
                                    plt.figure()
                                if filled:
                                    plt.fill_between(tmp[:, 0], sgn * tmp[:, 1], color=color, label=label)
                                else:
                                    plt.plot(tmp[:, 0],
                                             sgn * tmp[:, 1],
                                             marker + ls,
                                             color=color,
                                             label=label,
                                             lw=lw,
                                             ms=ms)
                        else:
                            if not nofig and sgn == 1:
                                plt.figure()
                            if color == '':
                                if lm == []:
                                    if filled:
                                        plt.fill_between(tmp[:, 0], sgn * tmp[:, 1:])
                                    else:
                                        plt.plot(tmp[:, 0], sgn * tmp[:, 1:], marker + ls, lw=lw, ms=ms)
                                else:
                                    for ilm in lm:
                                        lmname = label + ' '
                                        if ilm == 1:
                                            lmname += 's'
                                        if ilm == 2:
                                            lmname += 'p_x'
                                        if ilm == 3:
                                            lmname += 'p_y'
                                        if ilm == 4:
                                            lmname += 'p_z'
                                        if ilm == 5:
                                            lmname += 'd_{x^2-y^2}'
                                        if ilm == 6:
                                            lmname += 'd_{xz}'
                                        if ilm == 7:
                                            lmname += 'd_{z^2}'
                                        if ilm == 8:
                                            lmname += 'd_{yz}'
                                        if ilm == 9:
                                            lmname += 'd_{xy}'
                                        if ilm == 10:
                                            lmname += 'f_{-3}'
                                        if ilm == 11:
                                            lmname += 'f_{-2}'
                                        if ilm == 12:
                                            lmname += 'f_{-1}'
                                        if ilm == 13:
                                            lmname += 'f_{0}'
                                        if ilm == 14:
                                            lmname += 'f_{1}'
                                        if ilm == 15:
                                            lmname += 'f_{2}'
                                        if ilm == 16:
                                            lmname += 'f_{3}'
                                        plt.plot(tmp[:, 0],
                                                 sgn * tmp[:, 1 + ilm],
                                                 marker + ls,
                                                 lw=lw,
                                                 ms=ms,
                                                 label=lmname)
                            else:
                                if lm == []:
                                    if filled:
                                        plt.fill_between(tmp[:, 0], sgn * tmp[:, 1:], color=color)
                                    else:
                                        plt.plot(tmp[:, 0], sgn * tmp[:, 1:], marker + ls, color=color, lw=lw, ms=ms)
                                else:
                                    for ilm in lm:
                                        lmname = label + ' '
                                        if ilm == 1:
                                            lmname += 's'
                                        if ilm == 2:
                                            lmname += 'p_x'
                                        if ilm == 3:
                                            lmname += 'p_y'
                                        if ilm == 4:
                                            lmname += 'p_z'
                                        if ilm == 5:
                                            lmname += 'd_{x^2-y^2}'
                                        if ilm == 6:
                                            lmname += 'd_{xz}'
                                        if ilm == 7:
                                            lmname += 'd_{z^2}'
                                        if ilm == 8:
                                            lmname += 'd_{yz}'
                                        if ilm == 9:
                                            lmname += 'd_{xy}'
                                        if ilm == 10:
                                            lmname += 'f_{-3}'
                                        if ilm == 11:
                                            lmname += 'f_{-2}'
                                        if ilm == 12:
                                            lmname += 'f_{-1}'
                                        if ilm == 13:
                                            lmname += 'f_{0}'
                                        if ilm == 14:
                                            lmname += 'f_{1}'
                                        if ilm == 15:
                                            lmname += 'f_{2}'
                                        if ilm == 16:
                                            lmname += 'f_{3}'
                                        plt.plot(tmp[:, 0],
                                                 sgn * tmp[:, 1 + ilm],
                                                 marker + ls,
                                                 lw=lw,
                                                 ms=ms,
                                                 label=lmname,
                                                 color=color)
                            plt.title(label + ' ' + i)

                    #sum data
                    if first:
                        d = tmp
                        d[:, 1:] = d[:, 1:] * scale
                        first = False
                    else:
                        d[:, 1:] += tmp[:, 1:] * scale

    # set units and axis labels
    if 'eV' in units:
        ef = ef * 13.6
    if 'rel' in units:
        ef = 0

    #if lmdos:
    #   if lm<>[]:
    #      d = d[:,[0,1]+list(np.array(lm)+1)]
    #   d[:,1] = np.sum(d[:,2:], axis=1)
    if lm != []:
        d = d[:, [0, 1] + list(np.array(lm) + 1)]
    d[:, 1] = np.sum(d[:, 2:], axis=1)

    if normalized:
        d[:, 1:] = d[:, 1:] / (d[:, 1:]).max()

    xlab = r'E (Ry)'
    ylab = r'DOS (states/Ry)'
    if units == 'eV':
        xlab = r'E (eV)'
        ylab = r'DOS (states/eV)'
    elif units == 'eV_rel':
        xlab = r'E-E_F (eV)'
        ylab = r'DOS (states/eV)'
    elif units == 'Ry_rel':
        xlab = r'E-E_F (Ry)'
        ylab = r'DOS (states/Ry)'

    # plot dos
    if not allatoms:
        if totonly:
            if color == '':
                if xyswitch:
                    plt.plot(d[:, 1], d[:, 0] + shift, marker + ls, label=label, lw=lw, ms=ms)
                else:
                    plt.plot(d[:, 0] + shift, d[:, 1], marker + ls, label=label, lw=lw, ms=ms)
            else:
                if xyswitch:
                    if filled:
                        plt.fill_between(d[:, 1], d[:, 0] + shift, color=color, label=label)
                    else:
                        plt.plot(d[:, 1], d[:, 0] + shift, marker + ls, color=color, label=label, lw=lw, ms=ms)
                else:
                    if filled:
                        plt.fill_between(d[:, 0] + shift, d[:, 1], color=color, label=label)
                    else:
                        plt.plot(d[:, 0] + shift, d[:, 1], marker + ls, color=color, label=label, lw=lw, ms=ms)

        else:
            if color == '':
                if xyswitch:
                    plt.plot(d[:, 1:], d[:, 0] + shift, marker + ls, lw=lw, ms=ms)
                else:
                    plt.plot(d[:, 0] + shift, d[:, 1:], marker + ls, lw=lw, ms=ms)
            else:
                if xyswitch:
                    if filled:
                        plt.fill_between(d[:, 1], d[:, 0] + shift, color=color)
                    else:
                        plt.plot(d[:, 1:], d[:, 0] + shift, marker + ls, color=color, lw=lw, ms=ms)
                else:
                    if filled:
                        plt.fill_between(d[:, 0] + shift, d[:, 1], color=color)
                    else:
                        plt.plot(d[:, 0] + shift, d[:, 1:], marker + ls, color=color, lw=lw, ms=ms)
            plt.title(label)

    # plot fermi level
    if not noefline:
        if color != '' and efcolor == '':
            efcolor = color
        if efcolor == '':
            if xyswitch:
                plt.axhline(ef, ls=ls_ef, lw=lw_ef, color='grey')
            else:
                plt.axvline(ef, ls=ls_ef, lw=lw_ef, color='grey')
        else:
            if xyswitch:
                plt.axhline(ef, color=efcolor, ls=ls_ef, lw=lw_ef)
            else:
                plt.axvline(ef, color=efcolor, ls=ls_ef, lw=lw_ef)

    # set axis labels
    if xyswitch:
        plt.ylabel(xlab)
        plt.xlabel(ylab)
    else:
        plt.xlabel(xlab)
        plt.ylabel(ylab)

    if return_data:
        return d, ef



# -*- coding: utf-8 -*-
"""
Functions for expanding/splitting or converting electron configuration strings
"""
from masci_tools.util.constants import PERIODIC_TABLE_ELEMENTS

all_econfig = [
    '1s2', '2s2', '2p6', '3s2', '3p6', '4s2', '3d10', '4p6', '5s2', '4d10', '5p6', '6s2', '4f14', '5d10', '6p6', '7s2',
    '5f14', '6d10', '7p6', '8s2', '6f14'
]
states_spin = {'s': ['1/2'], 'p': ['1/2', '3/2'], 'd': ['3/2', '5/2'], 'f': ['5/2', '7/2']}
max_state_occ = {'s': 2., 'p': 6., 'd': 10., 'f': 14.}
max_state_occ_spin = {'1/2': 2., '3/2': 4., '5/2': 6., '7/2': 8.}
ATOMIC_NAMES = {data['symbol']: num for num, data in PERIODIC_TABLE_ELEMENTS.items()}


def get_econfig(element, full=False):
    """
    returns the econfiguration as a string of an element.

    :param element: element string
    :param full: a bool if True the econfig without [He]... is returned
    :returns: a econfig string
    """
    if isinstance(element, int):
        econ = PERIODIC_TABLE_ELEMENTS.get(element, {}).get('econfig')
    elif isinstance(element, str):
        element_num = ATOMIC_NAMES.get(element, None)
        econ = PERIODIC_TABLE_ELEMENTS.get(element_num, {}).get('econfig')
    else:
        raise ValueError('element has to be and int or string')

    if full and econ is not None:
        econ = rek_econ(econ)

    return econ


def get_coreconfig(element, full=False):
    """
    returns the econfiguration as a string of an element.

    :param element: element string
    :param full: a bool if True the econfig without [He]... is returned
    :return: coreconfig string
    """
    econ = get_econfig(element, full=full)
    return econ.split('|', maxsplit=1)[0].rstrip() if econ is not None else None


def rek_econ(econfigstr):
    """
    recursive routine to return a full econfig
    '[Xe] 4f14 | 5d10 6s2 6p4' -> '1s 2s ... 4f14 | 5d10 6s2 6p4'

    :param econfigstr: electron config string to expand

    :returns: expanded econfig string
    """
    split_econ = econfigstr.strip('[')
    split_econ = split_econ.split(']')
    if len(split_econ) == 1:
        return econfigstr
    else:
        rest = split_econ[1]
        elem = split_econ[0]
        econfig = get_econfig(elem)
        econ = econfig.replace(' |', '')
        econfigstr = rek_econ(econ + rest)
        return econfigstr  # for now


def convert_fleur_config_to_econfig(fleurconf_str, keep_spin=False):
    """
    '[Kr] (4d3/2) (4d5/2) (4f5/2) (4f7/2)' -> '[Kr] 4d10 4f14', or '[Kr] 4d3/2 4d5/2 4f5/2 4f7/2'

    # for now only use for coreconfig, it will fill all orbitals, since it has no information on the filling.

    :param fleurconf_str: string of the electron config like it is read from the inp.xml
    :param keep_spin: bool if True the spin indices will be kept in the converted string

    :returns: string of the electron config to be used in the inpgen
    """

    econfstring = fleurconf_str.replace('(', '').replace(')', '')

    if keep_spin:
        econfstring.split()
    else:
        elist = econfstring.split()
        econfstring_new = ''
        for state in elist:
            if '/' in state:
                # check if nl was added before if not add it with full occ
                base = state[:2]
                spin = state[2:]
                occ = max_state_occ_spin.get(spin)
                if base not in econfstring_new:
                    econfstring_new = f'{econfstring_new}{base}{int(occ)} '
                else:
                    max_occ = max_state_occ.get(base[1])
                    econfstring_new = econfstring_new.split(base, maxsplit=1)[0] + f'{base}{int(max_occ)} '
                    # we assume here that the two states come behind each other, ... rather bad
                    #econfstring_new.replace('{}'.format(base)
            else:
                econfstring_new = econfstring_new + state + ' '
        econfstring = econfstring_new

    return econfstring.strip()



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
#                                                                             #
###############################################################################
"""
This module contains a subclass of :py:class:`~masci_tools.vis.parameters.Plotter` for the matplotlib library
"""
from .parameters import Plotter, _generate_plot_parameters_table
import matplotlib.pyplot as plt
from matplotlib import cm
import copy


class MatplotlibPlotter(Plotter):
    """
    Class for plotting parameters and standard code snippets for plotting with the
    matplotlib backend.

    Kwargs in the __init__ method are forwarded to setting default values for the instance

    For specific documentation about the parameter/defaults handling refer to
    :py:class:`~masci_tools.vis.parameters.Plotter`.

    Below the current defined default values are shown:

    """
    _MATPLOTLIB_DEFAULTS = {
        # figure properties
        'title_fontsize': 16,
        'figure_kwargs': {
            'figsize': (8, 6),
            'dpi': 100,
            'facecolor': 'w',
            'edgecolor': 'k',
            'constrained_layout': False,
        },

        # axis properties
        'alpha': 1,
        'axis_linewidth': 1.5,
        'use_axis_formatter': False,
        'set_powerlimits': True,
        'xticks': None,
        'xticklabels': None,
        'yticks': None,
        'yticklabels': None,
        'invert_xaxis': False,
        'invert_yaxis': False,
        'color_cycle': None,
        'sub_colormap': None,

        # plot properties
        'linewidth': 2.0,
        'linestyle': '-',
        'marker': 'o',
        'markersize': 4.0,
        'color': None,
        'zorder': None,
        'repeat_colors_after': None,
        'edgecolor': None,
        'facecolor': None,
        'plot_label': None,
        'area_plot': False,
        'area_vertical': False,
        'area_enclosing_line': True,
        'area_alpha': 1.0,
        'area_linecolor': None,
        'plot_alpha': 1.0,
        'cmap': 'viridis',
        'norm': None,
        'shading': 'gouraud',
        'rasterized': True,

        #scale and limits placeholder
        'scale': None,
        'limits': None,

        # x, y label
        'labelfontsize': 15,
        'lines': None,
        'line_options': {
            'linestyle': '--',
            'color': 'k',
            'linewidth': 1.0
        },
        'font_options': {
            'family': 'serif',
            'color': 'black',
            'weight': 'normal',
            'size': 16,
        },

        # ticks
        'tick_paramsx': {
            'size': 4.0,
            'width': 1.0,
            'labelsize': 14,
            'length': 5,
            'labelrotation': 0
        },
        'tick_paramsy': {
            'size': 4.0,
            'width': 1.0,
            'labelsize': 14,
            'length': 5,
            'labelrotation': 0
        },
        'tick_paramsx_minor': {
            'size': 2.0,
            'width': 1.0,
            'labelsize': 0,
            'length': 2.5
        },
        'tick_paramsy_minor': {
            'size': 2.0,
            'width': 1.0,
            'labelsize': 0,
            'length': 2.5
        },
        'colorbar': True,
        'colorbar_padding': 0.1,
        # legend properties
        'legend': False,
        'legend_show_data_labels': False,
        'legend_remove_duplicates': False,
        'legend_options': {
            'fontsize': 'large',
            'linewidth': 3.0,
            'loc': 'best',
            'fancybox': True
        },

        # save all plots?
        'save_plots': False,  # True
        'save_format': 'png',  #'pdf'
        'save_options': {
            'transparent': True
        },
        'tightlayout': False,
        'show': True,
        # write data to file
        'save_raw_plot_data': False,
        'raw_plot_data_format': 'txt'
    }

    _MATPLOTLIB_DESCRIPTIONS = {
        # figure properties
        'title_fontsize':
        'Fontsize for the title of the figure',
        'figure_kwargs':
        'Arguments passed to `plt.figure` when creating the figure. '
        'Includes things like figsize, dpi, background color, ...',

        # axis properties
        'alpha':
        'Float specifying the transparency of the title',
        'axis_linewidth':
        'Linewidth of the lines for the axis',
        'use_axis_formatter':
        'If True the labels will always not be formatted '
        'with an additive constant at the top',
        'set_powerlimits':
        'If True the threshold for switching to scientific notation is adjusted to 0,3',
        'xticks':
        'Positions of the ticks on the x axis',
        'xticklabels':
        'Labels for the ticks on the x-axis',
        'yticks':
        'Positions for the ticks on the y-axis',
        'yticklabels':
        'Labels for the ticks on the y-axis',
        'invert_xaxis':
        'If True the direction of the x-axis is inverted',
        'invert_yaxis':
        'If True the direction of the y-axis is inverted',
        'color_cycle':
        'If set this will override the default color cycle of matplotlib. '
        'Can be given as name of a colormap cycle or list of colors',
        'sub_colormap':
        'If a colormap is used this can be used to cut out a part of the colormap. '
        'For example (0.5,1.0) will only use the upper half of the colormap',

        # plot properties
        'linewidth':
        'Linewidth for the plot(s)',
        'linestyle':
        'Linestyle for the plot(s)',
        'marker':
        'Shape of the marker to use for the plot(s)',
        'markersize':
        'Size of the markers to use in the plot(s)',
        'color':
        'Color to use in the plot(s)',
        'zorder':
        'z-position to use for the plot(s) (Is used to define fore- and background)',
        'repeat_colors_after':
        'If set the colors will be repeated after the given number of plots. '
        'Only implemented for multiple_scatterplots',
        'edgecolor':
        'Edgecolor to use in the plot(s)',
        'facecolor':
        'Facecolor to use in the plot(s)',
        'plot_label':
        'Label to use in the plot(s) for the legend',
        'area_plot':
        'If True fill_between(x) will be used to produce the plot(s)',
        'area_vertical':
        'Determines, whether to use fill_between or fill_betweenx for area plots',
        'area_enclosing_line':
        'If True a enclosing line will be drawn around the area',
        'area_alpha':
        'Transparency to use for the area in the area plot(s)',
        'area_linecolor':
        'Color for the enclosing line in the area plot(s)',
        'plot_alpha':
        'Transparency to use for the plot(s)',
        'cmap':
        'Colormap to use for scatter/pcolormesh or 3D plots',
        'norm':
        'If set this norm will be used to normalize data for the colormapping',
        'shading':
        'Shading to use for pcolormesh plots',
        'rasterized':
        'Rasterize the pcolormesh when drawing vector graphics.',

        #scale and limits placeholder
        'scale':
        "Dict specifying the scales of the axis, e.g {'y': 'log'}"
        'will create a logarithmic scale on the y-axis',
        'limits':
        "Dict specifying the limits of the axis, e.g {'x': (-5,5)}",

        # x, y label
        'labelfontsize':
        'Fontsize for the labels on the axis',
        'lines':
        'Dict specifying straight help-lines to draw. '
        "For example {'vertical': 0, 'horizontal': [-1,1]} will draw a vertical line at 0 "
        'and two horizontal at -1 and 1',
        'line_options':
        'Color, width, and more options for the help-lines',
        'font_options':
        'Default font options that can be used for text annotations',

        # ticks
        'tick_paramsx':
        'Parameters for major ticks on the x-axis (Size, fontsize, ...)',
        'tick_paramsy':
        'Parameters for major ticks on the y-axis (Size, fontsize, ...)',
        'tick_paramsx_minor':
        'Parameters for minor ticks on the x-axis (Size, fontsize, ...)',
        'tick_paramsy_minor':
        'Parameters for minor ticks on the y-axis (Size, fontsize, ...)',
        'colorbar':
        'If True and the function implements color mapping, a colorbar is shown',
        'colorbar_padding':
        'Specifies the space between plot and colorbar',
        # legend properties
        'legend':
        'If True a legend for the plot is shown',
        'legend_show_data_labels':
        'If True the column names from the data argument are shown if not overwritten',
        'legend_remove_duplicates':
        'If True duplicate legend labels are removed',
        'legend_options':
        'Parameters for displaying the legend (Fontsize, location, ...)',

        # save all plots?
        'save_plots':
        'if True the plots will be saved to file',
        'save_format':
        'Formats to save the plots to, can be single or list of formats',
        'save_options':
        'Additional options for saving the plots to file',
        'tightlayout':
        'If True the tight layout will be used (NOT IMPLEMENTED)',
        'show':
        'If True plt.show will be called at the end of the routine',
        # write data to file
        'save_raw_plot_data':
        'If True the data for the plot is saved to file (NOT IMPLEMENTED)',
        'raw_plot_data_format':
        'Format in which to save the data for the plot (NOT IMPLEMENTED)'
    }

    _MATPLOTLIB_GENERAL_ARGS = {
        'save_plots', 'save_format', 'tightlayout', 'save_raw_plot_data', 'raw_plot_data_format', 'show', 'legend',
        'legend_options', 'colorbar', 'colorbar_padding', 'tick_paramsy', 'tick_paramsx', 'tick_paramsy_minor',
        'tick_paramsx_minor', 'font_options', 'line_options', 'labelfontsize', 'lines', 'scale', 'limits', 'xticks',
        'xticklabels', 'yticks', 'yticklabels', 'figure_kwargs', 'title_font_size', 'repeat_colors_after',
        'color_cycle', 'sub_colormap', 'save_options'
    }

    #Sets of keys with special purposes

    _PLOT_KWARGS = {'linewidth', 'linestyle', 'marker', 'markersize', 'color', 'plot_label', 'plot_alpha', 'zorder'}
    _PLOT_KWARGS_AREA = {'area_linecolor', 'area_alpha', 'zorder'}
    _PLOT_KWARGS_COLORMESH = {
        'linewidth', 'linestyle', 'shading', 'rasterized', 'cmap', 'norm', 'edgecolor', 'facecolor', 'plot_label',
        'plot_alpha', 'zorder'
    }
    _PLOT_KWARGS_HIST = {
        'linewidth', 'linestyle', 'color', 'plot_label', 'plot_alpha', 'edgecolor', 'facecolor', 'zorder'
    }

    __doc__ = __doc__ + _generate_plot_parameters_table(_MATPLOTLIB_DEFAULTS, _MATPLOTLIB_DESCRIPTIONS)

    def __init__(self, **kwargs):
        super().__init__(self._MATPLOTLIB_DEFAULTS,
                         general_keys=self._MATPLOTLIB_GENERAL_ARGS,
                         key_descriptions=self._MATPLOTLIB_DESCRIPTIONS,
                         **kwargs)

    def plot_kwargs(self,
                    ignore=None,
                    extra_keys=None,
                    plot_type='default',
                    post_process=True,
                    list_of_dicts=True,
                    **kwargs):
        """
        Creates a dict or list of dicts (for multiple plots) with the defined parameters
        for the plotting calls fo matplotlib

        :param ignore: str or list of str (optional), defines keys to ignore in the creation of the dict
        :param extra_keys: optional set for addtional keys to retrieve
        :param post_process: bool, if True the parameters are cleaned up for inserting them directly into matplotlib plitting functions

        Kwargs are used to replace values by custom parameters:

        Example for using a custom markersize::

            p = MatplotlibPlotter()
            p.add_parameter('marker_custom', default_from='marker')
            p.plot_kwargs(marker='marker_custom')

        This code snippet will return the standard parameters for a plot, but the value
        for the marker will be taken from the key `marker_custom`
        """
        if plot_type == 'default':
            kwargs_keys = self._PLOT_KWARGS
        elif plot_type == 'colormesh':
            kwargs_keys = self._PLOT_KWARGS_COLORMESH
        elif plot_type == 'histogram':
            kwargs_keys = self._PLOT_KWARGS_HIST

        if self.single_plot:
            any_area = self['area_plot']
        else:
            any_area = any(self[('area_plot', indx)] for indx in range(self.num_plots))

        if any_area:
            kwargs_keys = kwargs_keys | self._PLOT_KWARGS_AREA

        if extra_keys is not None:
            kwargs_keys = kwargs_keys | extra_keys

        #Insert custom keys to retrieve
        kwargs_keys = kwargs_keys.copy()
        for key, replace_key in kwargs.items():
            kwargs_keys.remove(key)
            kwargs_keys.add(replace_key)

        plot_kwargs = self.get_multiple_kwargs(kwargs_keys, ignore=ignore)

        #Rename replaced keys back to standard names
        for key, replace_key in kwargs.items():
            custom_val = plot_kwargs.pop(replace_key, None)
            if custom_val is not None:
                plot_kwargs[key] = custom_val

        if not post_process:
            return plot_kwargs

        if 'plot_label' in plot_kwargs:
            plot_kwargs['label'] = plot_kwargs.pop('plot_label')

        if 'plot_alpha' in plot_kwargs:
            plot_kwargs['alpha'] = plot_kwargs.pop('plot_alpha')

        if 'cmap' in plot_kwargs and self['sub_colormap'] is not None:
            if not isinstance(self['sub_colormap'], (tuple, list)):
                raise ValueError('sub_colormap has to be a tuple of two numbers')

            if isinstance(plot_kwargs['cmap'], list):
                for indx, cmap in enumerate(plot_kwargs['cmap']):
                    if isinstance(cmap, str):
                        cmap = plt.get_cmap(cmap)

                    if isinstance(self['sub_colormap'], list):
                        limits = self['sub_colormap'][indx]
                    else:
                        limits = self['sub_colormap']

                    plot_kwargs['cmap'][indx] = self.truncate_colormap(cmap, *limits)
            else:
                if isinstance(plot_kwargs['cmap'], str):
                    plot_kwargs['cmap'] = plt.get_cmap(plot_kwargs['cmap'])

                plot_kwargs['cmap'] = self.truncate_colormap(plot_kwargs['cmap'], *self['sub_colormap'])

        if list_of_dicts:
            plot_kwargs = self.dict_of_lists_to_list_of_dicts(plot_kwargs, self.single_plot, self.num_plots)

        if not list_of_dicts and 'label' in plot_kwargs:
            label = plot_kwargs.pop('label')
            if isinstance(label, list):
                plot_kwargs['label'] = [value if value is not None else '' for value in label]
            else:
                plot_kwargs['label'] = label if label is not None else ''

        if not self.single_plot and list_of_dicts:
            for index, value in enumerate(plot_kwargs):
                if self[('area_plot', index)]:
                    value.pop('marker', None)
                    value.pop('markersize', None)
                    value['alpha'] = value.pop('area_alpha')
                else:
                    value.pop('area_alpha', None)
                    value.pop('area_linecolor', None)
        else:
            if self['area_plot']:
                plot_kwargs.pop('marker', None)
                plot_kwargs.pop('markersize', None)
                plot_kwargs['alpha'] = plot_kwargs.pop('area_alpha')
            else:
                plot_kwargs.pop('area_alpha', None)
                plot_kwargs.pop('area_linecolor', None)

        if self['legend_show_data_labels']:
            if not self.single_plot and list_of_dicts:
                for index, value in enumerate(plot_kwargs):
                    if value.get('label') is None:
                        value.pop('label', None)
            elif plot_kwargs.get('label') is None:
                plot_kwargs.pop('label', None)
        else:
            if not self.single_plot and list_of_dicts:
                for index, value in enumerate(plot_kwargs):
                    if 'label' not in value:
                        value['label'] = None
            elif 'label' not in plot_kwargs:
                plot_kwargs['label'] = None

        return plot_kwargs

    def prepare_plot(self, title=None, xlabel=None, ylabel=None, zlabel=None, axis=None, minor=False, projection=None):
        """
        Prepares the figure of a matplotlib plot, setting the labels/titles, ticks, ...

        :param title: str for the title of the figure
        :param xlabel: str for the label on the x-axis
        :param ylabel: str for the label on the y-axis
        :param zlabel: str for the label on the z-axis
        :param axis: matplotlib axes object, optional, if given the operations are performed on the object
                     otherwise a new figure and subplot are created
        :param minor: bool, if True minor tick parameters are set
        :param projection: str, passed on to the add_subplot call

        :returns: the created or modified axis object
        """
        from cycler import cycler, Cycler

        if axis is not None:
            ax = axis
        else:
            fig = plt.figure(num=None, **self['figure_kwargs'])
            ax = fig.add_subplot(111, projection=projection)

        for axes in ['top', 'bottom', 'left', 'right']:
            ax.spines[axes].set_linewidth(self['axis_linewidth'])
        ax.set_title(title, fontsize=self['title_fontsize'], alpha=self['alpha'], ha='center')
        ax.set_xlabel(xlabel, fontsize=self['labelfontsize'])
        ax.set_ylabel(ylabel, fontsize=self['labelfontsize'])
        if zlabel is not None:
            ax.set_zlabel(zlabel, fontsize=self['labelfontsize'])
        ax.yaxis.set_tick_params(**self['tick_paramsy'])
        ax.xaxis.set_tick_params(**self['tick_paramsx'])

        if minor:
            ax.yaxis.set_tick_params(which='minor', **self['tick_paramsy_minor'])
            ax.xaxis.set_tick_params(which='minor', **self['tick_paramsx_minor'])

        if self['xticks'] is not None:
            ax.xaxis.set_ticks(self['xticks'])
        if self['xticklabels'] is not None:
            ax.xaxis.set_ticklabels(self['xticklabels'])

        if self['yticks'] is not None:
            ax.yaxis.set_ticks(self['yticks'])
        if self['yticklabels'] is not None:
            ax.yaxis.set_ticklabels(self['yticklabels'])

        if self['use_axis_formatter']:
            if self['set_powerlimits']:
                ax.yaxis.get_major_formatter().set_powerlimits((0, 3))
                ax.xaxis.get_major_formatter().set_powerlimits((0, 3))
            ax.yaxis.get_major_formatter().set_useOffset(False)
            ax.xaxis.get_major_formatter().set_useOffset(False)

        if self['color_cycle'] is not None:
            if isinstance(self['color_cycle'], str):
                ax.set_prop_cycle(cycler(color=plt.get_cmap(self['color_cycle']).colors))
            elif isinstance(self['color_cycle'], Cycler):
                ax.set_prop_cycle(self['color_cycle'])
            else:
                ax.set_prop_cycle(cycler(color=self['color_cycle']))

        if self['invert_xaxis']:
            ax.invert_xaxis()

        if self['invert_yaxis']:
            ax.invert_yaxis()

        return ax

    def set_scale(self, ax):
        """
        Set scale of the axis (for example 'log')

        :param ax: Axes object on which to perform the operation
        """

        if self['scale'] is not None:
            if 'x' in self['scale']:
                ax.set_xscale(self['scale']['x'])
            if 'y' in self['scale']:
                ax.set_yscale(self['scale']['y'])
            if 'z' in self['scale']:
                ax.set_zscale(self['scale']['z'])

    def set_limits(self, ax):
        """
        Set limits of the axis

        :param ax: Axes object on which to perform the operation
        """

        if self['limits'] is not None:
            if 'x' in self['limits']:
                xmin = self['limits']['x'][0]
                xmax = self['limits']['x'][1]
                if self['invert_xaxis']:
                    ax.set_xlim(xmax, xmin)
                else:
                    ax.set_xlim(xmin, xmax)
            if 'y' in self['limits']:
                ymin = self['limits']['y'][0]
                ymax = self['limits']['y'][1]
                if self['invert_yaxis']:
                    ax.set_ylim(ymax, ymin)
                else:
                    ax.set_ylim(ymin, ymax)
            if 'z' in self['limits']:
                zmin = self['limits']['z'][0]
                zmax = self['limits']['z'][1]
                ax.set_zlim(zmin, zmax)

    def draw_lines(self, ax):
        """
        Draw horizontal and vertical lines specified in the lines argument

        :param ax: Axes object on which to perform the operation
        """
        if self['lines'] is not None:
            if 'horizontal' in self['lines']:
                lines = copy.deepcopy(self['lines']['horizontal'])
                if not isinstance(lines, list):
                    lines = [lines]

                for line_def in lines:
                    options = copy.deepcopy(self['line_options'])
                    if isinstance(line_def, dict):
                        positions = line_def.pop('pos')
                        if not isinstance(positions, list):
                            positions = [positions]
                        options.update(line_def)
                    elif isinstance(line_def, list):
                        positions = line_def
                    else:
                        positions = [line_def]

                    for pos in positions:
                        ax.axhline(pos, **options)
            if 'vertical' in self['lines']:
                lines = copy.deepcopy(self['lines']['vertical'])
                if not isinstance(lines, list):
                    lines = [lines]

                for line_def in lines:
                    options = copy.deepcopy(self['line_options'])
                    if isinstance(line_def, dict):
                        positions = line_def.pop('pos')
                        if not isinstance(positions, list):
                            positions = [positions]
                        options.update(line_def)
                    elif isinstance(line_def, list):
                        positions = line_def
                    else:
                        positions = [line_def]

                    for pos in positions:
                        ax.axvline(pos, **options)

    def show_legend(self, ax, leg_elems=None):
        """
        Print a legend for the plot

        :param ax: Axes object on which to perform the operation
        """
        if leg_elems is None:
            if self['legend_remove_duplicates']:
                leg_elems = ax.get_legend_handles_labels()
                by_label = dict(zip(*reversed(leg_elems)))
                leg_elems = (by_label.values(), by_label.keys())
            else:
                leg_elems = ()

        if self['legend']:
            loptions = copy.deepcopy(self['legend_options'])
            linewidth = loptions.pop('linewidth', 1.5)
            title_font_size = loptions.get('fontsize', 15)
            leg = ax.legend(*leg_elems, **loptions)
            leg.get_frame().set_linewidth(linewidth)
            leg.get_title().set_fontsize(title_font_size)  #legend 'Title' fontsize

    def show_colorbar(self, ax):
        """
        Print a colorbar for the plot

        :param ax: Axes object on which to perform the operation
        """

        if self['colorbar']:
            if isinstance(self['cmap'], list):
                raise ValueError('show_colorbar only available for single colormaps')
            mappable = cm.ScalarMappable(cmap=self['cmap'], norm=self['norm'])
            if self['limits'] is not None:
                if 'color' in self['limits']:
                    cmin = self['limits']['color'][0]
                    cmax = self['limits']['color'][1]
                    mappable.set_clim(cmin, cmax)

            plt.colorbar(mappable, ax=ax, pad=self['colorbar_padding'])

    @staticmethod
    def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=256):
        """
        Cut off parts of colormap

        :param cmap: cmap to truncate
        :param minval: minimum value of new colormap
        :param maxval: maximum value of new colormap
        :param n: number of colors in new colormap

        :returns: colormap truncated to only hold colors between minval and maxval from old colormap
        """
        from matplotlib import colors
        import numpy as np

        new_cmap = colors.LinearSegmentedColormap.from_list(f'trunc({cmap.name},{minval:.2f},{maxval:.2f})',
                                                            cmap(np.linspace(minval, maxval, n)))

        return new_cmap

    def save_plot(self, saveas):
        """
        Save the current figure or show the current figure

        :param saveas: str, filename for the resulting file
        """
        if self['save_plots']:
            if isinstance(self['save_format'], list):
                formats = self['save_format']
            else:
                formats = [self['save_format']]

            for save_format in formats:
                savefilename = f'{saveas}.{save_format}'
                print(f'Save plot to: {savefilename}')
                plt.savefig(savefilename, format=save_format, **self['save_options'])
        if self['show']:
            plt.show()



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains functions to load an fleur out.xml file, parse it with a schema
and convert its content to a dict, based on the tasks given
"""
from masci_tools.util.parse_tasks import ParseTasks
from masci_tools.util.schema_dict_util import tag_exists, read_constants, eval_simple_xpath, evaluate_attribute
from masci_tools.util.xml.common_functions import clear_xml, validate_xml
from masci_tools.io.io_fleurxml import load_outxml
from masci_tools.util.logging_util import DictHandler, OutParserLogAdapter
from lxml import etree
import copy
import warnings
import logging


def outxml_parser(outxmlfile,
                  version=None,
                  parser_info_out=None,
                  iteration_to_parse=None,
                  strict=False,
                  debug=False,
                  base_url=None,
                  **kwargs):
    """
    Parses the out.xml file to a dictionary based on the version and the given tasks

    :param outxmlfile: either path to the out.xml file, opened file handle or a xml etree to be parsed
    :param version: version string to enforce that a given schema is used
    :param parser_info_out: dict, with warnings, info, errors, ...
    :param iteration_to_parse: either str or int, (optional, default 'last')
                               determines which iteration should be parsed.
                               Accepted are 'all', 'first', 'last' or an index for the iteration
    :param strict: bool if True  and no parser_info_out is provided any encountered error will immediately be raised
    :param debug: bool if True additional information is printed out in the logs

    Kwargs:
        :param ignore_validation: bool, if True schema validation errors are only logged
        :param minimal_mode: bool, if True only total Energy, iteration number and distances are parsed
        :param list_return: bool, if True one-item lists in the output dict are not converted to simple values
        :param additional_tasks: dict to define custom parsing tasks. For detailed explanation
                                 See :py:mod:`~masci_tools.io.parsers.fleur.default_parse_tasks`.
        :param overwrite: bool, if True and keys in additional_tasks collide with defaults
                          The defaults will be overwritten
        :param append: bool, if True and keys in additional_tasks collide with defaults
                       The inner tasks will be written into the dict. If inner keys collide
                       they are overwritten

    :return: python dictionary with the information parsed from the out.xml

    :raises ValueError: If the validation against the schema failed, or an irrecoverable error
                        occured during parsing
    :raises FileNotFoundError: If no Schema file for the given version was found
    :raises KeyError: If an unknown task is encountered

    """

    __parser_version__ = '0.5.0'

    logger = logging.getLogger(__name__)

    parser_log_handler = None
    if parser_info_out is not None or not strict:
        if parser_info_out is None:
            parser_info_out = {}

        logging_level = logging.INFO
        if debug:
            logging_level = logging.DEBUG
        logger.setLevel(logging_level)

        parser_log_handler = DictHandler(parser_info_out,
                                         WARNING='parser_warnings',
                                         ERROR='parser_errors',
                                         INFO='parser_info',
                                         DEBUG='parser_debug',
                                         CRITICAL='parser_critical',
                                         ignore_unknown_levels=True,
                                         level=logging_level)

        logger.addHandler(parser_log_handler)

    if strict:
        logger = None

    if logger is not None:
        logger.info('Masci-Tools Fleur out.xml Parser v%s', __parser_version__)

    outfile_broken = False
    try:
        xmltree, outschema_dict = load_outxml(outxmlfile, logger=logger, base_url=base_url, recover=False)
    except ValueError as err:
        if 'Failed to parse output file' in str(err):
            outfile_broken = True
            if logger is None:
                warnings.warn('The out.xml file is broken I try to repair it.')
            else:
                logger.warning('The out.xml file is broken I try to repair it.')
        else:
            if logger is not None:
                logger.error(str(err))
            raise

    if outfile_broken:
        try:
            xmltree, outschema_dict = load_outxml(outxmlfile, logger=logger, base_url=base_url, recover=True)
        except ValueError as err:
            if 'Failed to parse output file' in str(err):
                if logger is None:
                    raise ValueError('Skipping the parsing of the xml file. Repairing was not possible.') from err
                else:
                    logger.exception('Skipping the parsing of the xml file. ' 'Repairing was not possible.')
                    return {}
            if logger is not None:
                logger.error(str(err))
            raise

    actual_out_version = evaluate_attribute(xmltree, outschema_dict, 'fleurOutputVersion', logger=logger)
    if actual_out_version == '0.27':
        actual_inp_version = actual_out_version
    else:
        actual_inp_version = evaluate_attribute(xmltree, outschema_dict, 'fleurInputVersion', logger=logger)

    ignore_validation = False
    out_version = actual_out_version
    inp_version = actual_inp_version
    if outschema_dict['out_version'] != actual_out_version or \
       outschema_dict['inp_version'] != actual_inp_version:
        ignore_validation = True
        out_version = outschema_dict['out_version']
        inp_version = outschema_dict['inp_version']

    ignore_validation = kwargs.get('ignore_validation', ignore_validation)

    if logger is not None:
        logger.info('Found fleur out file with the versions out: %s; inp: %s', out_version, inp_version)

    xmltree, _ = clear_xml(xmltree)
    root = xmltree.getroot()

    errmsg = ''
    try:
        validate_xml(xmltree, outschema_dict.xmlschema, error_header='Output file does not validate against the schema')
    except etree.DocumentInvalid as err:
        errmsg = str(err)
        if logger is not None:
            logger.warning(errmsg)
        if not ignore_validation:
            if logger is not None:
                logger.exception(errmsg)
            raise ValueError(errmsg) from err

    if not outschema_dict.xmlschema.validate(xmltree) and errmsg == '':
        msg = 'Output file does not validate against the schema: Reason is unknown'
        if logger is not None:
            logger.warning(msg)
        if not ignore_validation:
            if logger is not None:
                logger.exception(msg)
            raise ValueError(msg)

    parser = ParseTasks(out_version)
    additional_tasks = kwargs.pop('additional_tasks', {})
    for task_name, task_definition in additional_tasks.items():
        parser.add_task(task_name, task_definition, **kwargs)

    out_dict, constants = parse_general_information(root,
                                                    parser,
                                                    outschema_dict,
                                                    logger=logger,
                                                    iteration_to_parse=iteration_to_parse,
                                                    **kwargs)

    out_dict['input_file_version'] = outschema_dict['inp_version']
    # get all iterations in out.xml file
    iteration_nodes = eval_simple_xpath(root, outschema_dict, 'iteration', logger=logger, list_return=True)
    n_iters = len(iteration_nodes)

    # parse only last stable interation
    # (if modes (dos and co) maybe parse anyway if broken?)
    if outfile_broken and (n_iters >= 2):
        iteration_nodes = iteration_nodes[:-2]
        if logger is not None:
            logger.info('The last parsed iteration is %s', n_iters - 2)
    elif outfile_broken and (n_iters == 1):
        iteration_nodes = [iteration_nodes[0]]
        if logger is not None:
            logger.info('The last parsed iteration is %s', n_iters)
    elif not outfile_broken and (n_iters >= 1):
        pass
    else:  # there was no iteration found.
        # only the starting charge density could be generated
        msg = 'There was no iteration found in the outfile, either just a ' \
              'starting density was generated or something went wrong.'
        if logger is None:
            raise ValueError(msg)
        else:
            logger.error(msg)

    if iteration_to_parse is None:
        iteration_to_parse = 'last'  #This is the default from the aiida_fleur parser

    if iteration_to_parse == 'last':
        iteration_nodes = iteration_nodes[-1]
    elif iteration_to_parse == 'first':
        iteration_nodes = iteration_nodes[0]
    elif iteration_to_parse == 'all':
        pass
    elif isinstance(iteration_to_parse, int):
        try:
            iteration_nodes = iteration_nodes[iteration_to_parse]
        except IndexError as exc:
            if logger is not None:
                logger.exception(exc)
            raise ValueError(f"Invalid value for iteration_to_parse: Got '{iteration_to_parse}'"
                             f"; but only '{len(iteration_nodes)}' iterations are available") from exc
    else:
        if logger is not None:
            logger.error(
                "Invalid value for iteration_to_parse: Got '%s' "
                "Valid values are: 'first', 'last', 'all', or int", iteration_to_parse)
        raise ValueError(f"Invalid value for iteration_to_parse: Got '{iteration_to_parse}' "
                         "Valid values are: 'first', 'last', 'all', or int")

    if not isinstance(iteration_nodes, list):
        iteration_nodes = [iteration_nodes]

    logger_info = {'iteration': 'unknown'}
    iteration_logger = OutParserLogAdapter(logger, logger_info)

    for node in iteration_nodes:
        iteration_number = evaluate_attribute(node, outschema_dict, 'numberForCurrentRun', optional=True)

        if iteration_number is not None:
            logger_info['iteration'] = iteration_number

        out_dict = parse_iteration(node, parser, outschema_dict, out_dict, constants, logger=iteration_logger, **kwargs)

        logger_info['iteration'] = 'unknown'

    if not kwargs.get('list_return', False):
        #Convert one item lists to simple values
        for key, value in out_dict.items():
            if isinstance(value, list):
                if len(value) == 1:
                    out_dict[key] = value[0]
            elif isinstance(value, dict):
                for subkey, subvalue in value.items():
                    if isinstance(subvalue, list):
                        if len(subvalue) == 1:
                            out_dict[key][subkey] = subvalue[0]

    if parser_log_handler is not None:
        if logger is not None:
            logger.removeHandler(parser_log_handler)

    return out_dict


def parse_general_information(root, parser, outschema_dict, logger, iteration_to_parse=None, **kwargs):
    """
    Parses the information from the out.xml outside scf iterations

    Also defined constants and fleur calculation modes are read in

    Args:
        :param root: etree Element for the root of the out.xml
        :param parser: ParseTasks object with all defined tasks
        :param outschema_dict: dict with the information parsed from the OutputSchema
        :param parser_info_out: dict, with warnings, info, errors, ...

    Kwargs:
        :param minimal_mode: bool, if True only total Energy, iteration number and distances are parsed

    """
    from masci_tools.util.xml.xml_getters import get_fleur_modes

    minimal_mode = kwargs.get('minimal_mode', False)
    if iteration_to_parse is None:
        iteration_to_parse = 'last'

    constants = read_constants(root, outschema_dict, logger=logger)
    if logger is not None:
        logger.info('The following defined constants were found: %s', constants)

    fleurmode = get_fleur_modes(root, outschema_dict, logger=logger)
    if logger is not None:
        logger.info('The following Fleur modes were found: %s', fleurmode)

    parser.determine_tasks(fleurmode, minimal=minimal_mode)

    #For certain fleur modes we need to overwrite the tasks
    if fleurmode['dos'] or fleurmode['band'] or fleurmode['cf_coeff']:
        parser.iteration_tasks = ['iteration_number', 'fermi_energy']
        if fleurmode['bz_integration'] == 'hist':
            parser.iteration_tasks = ['iteration_number', 'fermi_energy', 'bandgap']

    if fleurmode['plot']:
        parser.iteration_tasks = []  #In this case there are multiple possibilities where fleur terminates
        #So we discard all the iteration tasks

    if fleurmode['relax'] and iteration_to_parse == 'last':
        if 'distances' in parser.iteration_tasks:
            parser.iteration_tasks.remove('distances')
        if 'magnetic_distances' in parser.iteration_tasks:
            parser.iteration_tasks.remove('magnetic_distances')
        if 'nmmp_distances' in parser.iteration_tasks:
            parser.iteration_tasks.remove('nmmp_distances')

    if logger is not None:
        logger.debug('The following tasks are performed on the root: %s', parser.general_tasks)

    out_dict = {'fleur_modes': fleurmode}

    for task in parser.general_tasks:

        if logger is not None:
            logger.debug('Performing task: %s', task)
        out_dict = parser.perform_task(task, root, out_dict, outschema_dict, constants, logger=logger, use_lists=False)

    return out_dict, constants


def parse_iteration(iteration_node, parser, outschema_dict, out_dict, constants, logger, **kwargs):
    """
    Parses an scf iteration node. Which tasks to perform is stored in parser.iteration_tasks

    Args:
        :param iteration_node: etree Element for a scf iteration
        :param parser: ParseTasks object with all defined tasks
        :param outschema_dict: dict with the information parsed form the OutputSchema
        :param out_dict: dict with the parsed results
        :param constants: dict with all the defined mathematical constants
        :param parser_info_out: dict, with warnings, info, errors, ...

    Kwargs:
        :param strict: bool, if True an error will be raised if an unknown task is encountered
                       otherwise a warning is written to parser_info_out
        :param minimal_mode: bool, if True only total Energy, iteration number and distances are parsed
    """

    minimal_mode = kwargs.get('minimal_mode', False)

    iteration_tasks = copy.deepcopy(parser.iteration_tasks)
    #If the iteration is a forcetheorem calculation
    #Replace all tasks with the given tasks for the calculation
    forcetheorem_tags = ['Forcetheorem_DMI', 'Forcetheorem_SSDISP', 'Forcetheorem_JIJ', 'Forcetheorem_MAE']
    for tag in forcetheorem_tags:
        exists = tag_exists(iteration_node, outschema_dict, tag)
        if exists:
            if minimal_mode:
                iteration_tasks = []
            else:
                iteration_tasks = [tag.lower()]
            break

    if logger is not None:
        logger.debug('The following tasks are performed for the iteration: %s', iteration_tasks)

    for task in iteration_tasks:

        if logger is not None:
            logger.debug('Performing task: %s', task)

        try:
            out_dict = parser.perform_task(task, iteration_node, out_dict, outschema_dict, constants, logger=logger)
        except KeyError:
            if logger is not None:
                logger.exception("Unknown task: '%s'. Skipping this one", task)
            raise

    return out_dict



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), 2018 Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.    #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools    #
# For further information on the license, see the LICENSE.txt file            #
#                                                                             #
###############################################################################
# by David Bauer 2012
# edited by Philipp Ruessmann 2014
# added averaging of spin up/down by Philipp Sep. 2015
# pylint: skip-file
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from numpy import *
from sys import argv, exit
from six.moves import map
from six.moves import range
from six.moves import input

mode = 'pot'
if len(argv) == 2:
    if argv[1] == 'shape' or argv[1] == 'shapefun':
        mode = 'shape'

if mode == 'pot':
    data = open('potential').readlines()
elif mode == 'shape':
    data = open('shapefun').readlines()


def iatom2shape():
    atominfolines = open('inputcard').readlines()
    print(len(atominfolines))
    starti = -1
    for i in range(len(atominfolines)):
        if 'ATOMINFO' in atominfolines[i]:
            starti = i
    icount = 2
    listatom = []
    while (not ('---' in atominfolines[starti + icount])):
        print()
        listatom.append(int(atominfolines[starti + icount].split()[8]))
        icount = icount + 1
    return listatom


def iatom2shape_2016():
    inputcard = open('inputcard').readlines()
    try:
        natyp = [int(line.split('=')[1].split()[0]) for line in inputcard if 'NAEZ' in line][0]
        print('natyp v1:', natyp)
    except:
        natyp = [int(line.split('=')[1].split()[0]) for line in inputcard if 'NATYP' in line][0]
        print('natyp v2:', natyp)
    try:
        ntc_iline = [iline for iline in range(len(inputcard)) if '<SHAPE>' in inputcard[iline]][0]
        ntc_line_start = inputcard[ntc_iline].find('<SHAPE>')
    except:
        ntc_iline = [iline for iline in range(len(inputcard)) if 'NTC' in inputcard[iline]][0]
        ntc_line_start = inputcard[ntc_iline].find('NTC')
    listatom = []
    for iatom in range(natyp):
        line = inputcard[ntc_iline + iatom + 1]
        ntc = int(line[ntc_line_start:].split()[0])
        listatom.append(ntc)
    return listatom


def check_potstart(str1, shape_ver='new'):
    if mode == 'pot':
        #check1='POTENTIAL' in str1
        check1 = 'exc:' in str1
    elif mode == 'shape':
        if shape_ver == 'new':
            check1 = 'Shape number' in str1
        else:
            check1 = (len(str1) == 11)
    else:
        print('mode error')
        exit()
    return check1


def read_pot_values(lstart, lstop):
    tmpdat = data[lstart:lstop]
    #print shape(tmpdat)
    #first find start of potential values
    il, lpotstart = 0, -1
    for i in range(25):
        tmplen = len(tmpdat[i])
        #print tmplen
        if tmplen == 81 and lpotstart == -1:
            lpotstart = il
        il += 1
    #now lpotstart contains position where values begin
    #next step is to read in values which can later be modified
    potential = []
    potentialheader = tmpdat[:lpotstart]
    lm = 0
    for il in range(lpotstart, len(tmpdat)):
        s = tmpdat[il]
        s = s.replace('D', 'E')
        snew = []
        if il == lpotstart:
            dl = len(s) // 4
            print(dl)
            print(len(s))
        snew.append(s[:dl])
        if len(s) > 21:
            snew.append(s[dl:2 * dl])
        if len(s) > 41:
            snew.append(s[2 * dl:3 * dl])
        if len(s) > 61:
            snew.append(s[3 * dl:-1])
        print(snew)
        potential.append(snew)
    #now array potential contains potential values as float numbers
    return potentialheader, potential


def pot2newdata(header, pot):
    """ convert potential data back to string file format """
    out = header
    for i in pot:
        if len(i) == 1:
            out.append(i[0])
        else:
            tmps = ''
            for ii in i:
                tmps += ii.replace('E', 'D').replace('e', 'D')
            tmps += '\n'
            out.append(tmps)
    return out


def calc_coulomb_spin_pot(pot1, pot2):
    """ calculate spin mixing and coulomb part of potential """
    # first do consistency checks
    if len(pot1) != len(pot2):
        exit('ERROR: potential1 and potential2 inconsistent in calc_coulomb_spin_pot')
    # now calculate coulomb part and spin part and save them in vc and vs
    vc, vs = [], []
    for i in range(len(pot1)):
        if len(pot1) != len(pot2):
            exit('ERROR: potential1 and potential2 inconsistent in calc_coulomb_spin_pot')
        tmp1, tmp2 = pot1[i], pot2[i]
        tmpvc, tmpvs = [], []
        if len(tmp1[0]) >= 20:
            for ii in range(len(tmp1)):
                tmpvc.append('%20.13e' % ((float(tmp1[ii]) + float(tmp2[ii])) / 2.))
                tmpvs.append('%20.13e' % ((float(tmp1[ii]) - float(tmp2[ii])) / 2.))
        else:
            tmpvc.append(tmp1[0])
            tmpvs.append(tmp1[0])
        vc.append(tmpvc)
        vs.append(tmpvs)
    # now vc and vs contain spin and coulomb part of potential
    return vc, vs


def combine_potentials(pot1, pot2, alpha):
    """ combine potentials: pot_out = pot1 + alpha*pot2 """
    # first do consistency checks
    if len(pot1) != len(pot2):
        exit('ERROR: potential1 and potential2 inconsistent in combine_potentials')
    # now calculate coulomb part and spin part and save them in vc and vs
    potout = []
    for i in range(len(pot1)):
        if len(pot1) != len(pot2):
            exit('ERROR: potential1 and potential2 inconsistent in combine_potentials')
        tmp1, tmp2 = pot1[i], pot2[i]
        tmppot = []
        if len(tmp1[0]) >= 20:
            for ii in range(len(tmp1)):
                tmppot.append('%20.13e' % (float(tmp1[ii]) + alpha * float(tmp2[ii])))
        else:
            tmppot.append(tmp1[0])
        potout.append(tmppot)
    return potout


def combine_header(head1, head2, alpha):
    print(len(head1), len(head2), alpha)
    # combine core energies
    headnew1 = head1[:7]
    ne = int(head1[6].split()[0])
    ne2 = int(head2[6].split()[0])
    if ne != ne2:
        exit('ERROR: number of core levels inconsistent')
    for i in range(ne):
        tmp1 = head1[7 + i].replace('D', 'E').split()
        tmp2 = head2[7 + i].replace('D', 'E').split()
        tmp = ('%5i%20.11e\n' % (int(tmp1[0]), ((float(tmp1[1]) + float(tmp2[1])) / 2. + alpha *
                                                ((float(tmp1[1]) - float(tmp2[1])) / 2.)))).replace('e', 'D')
        headnew1.append(tmp)
    print('rest length', len(head1[7 + ne:]))
    for i in range(len(head1[7 + ne:])):
        headnew1.append(head1[7 + ne:][i])
    print(ne)
    return headnew1


index1 = []
index2 = []
for i in range(len(data)):
    if check_potstart(data[i]):
        index1.append(i)
        if len(index1) > 1:
            index2.append(i - 1)
index2.append(i)

# read shapefun if old style is used
if mode == 'shape' and len(index1) < 1:
    index1 = []
    index2 = []
    for i in range(len(data)):
        if check_potstart(data[i], shape_ver='old'):
            index1.append(i)
            if len(index1) > 1:
                index2.append(i - 1)
    index2.append(i)

print(index1)
print(index2)

print('Potential file read')
print('found %i potentials in file' % len(index1))
print('')

tempsave = -1
order = list(range(len(index1)))
while 1:
    #while 0:
    for i in range(len(order)):
        #print '%3i'%i,data[index1[order[i]]][:-1]
        if 'GENERAL POTENTIAL' in data[index1[order[i]]][:-1]:
            print('%3i' % i, data[index1[order[i]]][:-1], data[index1[order[i]] + 1][:-1])
        else:
            print('%3i' % i, data[index1[order[i]]][:-1])
    print('  ***************************')
    print('  *  ( 0) Stop and Save         *')
    print('  *  ( 1) Delete                *')
    print('  *  ( 2) Copy                  *')
    print('  *  ( 3) Cut                   *')
    print('  *  ( 4) Paste                 *')
    print('  *  ( 5) Copy  [multiline]     *')
    print('  *  ( 6) Cut   [multiline]     *')
    print('  *  ( 7) Del   [multiline]     *')
    print('  *  ( 8) Paste [multiple]      *')
    print('  *  ( 9) swap lines            *')
    print('  *  (10) read scoef nonmag host*')
    print('  *  (11) flip magnetic moments *')
    print('  *  (12) Input new list        *')
    print('  *  (13) Nspin=1 -> 2  :       *')
    print('  *  (14) read scoef mag host   *')
    print('  *  (15) average spin up/down  *')
    print('  *  (16) scale up mag. moment  *')
    print('  ***************************')

    mode1 = int(eval(input('Input: ')))
    print(mode1)
    if mode1 == 0:
        break
    if mode1 == 1:
        row1 = int(eval(input('Row number:')))
        del order[row1]
    if mode1 == 2:
        row1 = int(eval(input('Row number:')))
        tempsave = order[row1]
    if mode1 == 3:
        row1 = int(eval(input('Row number:')))
        tempsave = order[row1]
        del order[row1]
    if mode1 == 4:
        if type(tempsave) == type(1):
            if tempsave != -1:
                row1 = int(eval(input('Paste before number:')))
                order.insert(row1, tempsave)
            else:
                print('nothing in temp, copy first')
                eval(input('Continue'))
        else:
            row1 = int(eval(input('Paste before number:')))
            print(row1)
            print(tempsave)
            for i in reversed(list(range(len(tempsave)))):
                order.insert(row1, tempsave[i])
    if mode1 == 5:
        row1 = int(eval(input('Row number [start]:')))
        row2 = int(eval(input('Row number [stop] :')))
        tempsave = list(range(row1, row2 + 1))
    if mode1 == 6:
        row1 = int(eval(input('Row number [start]:')))
        row2 = int(eval(input('Row number [stop] :')))
        tempsave = list(range(row1, row2 + 1))
        for i in range(len(tempsave)):
            del order[row1]
    if mode1 == 7:
        row1 = int(eval(input('Row number [start]:')))
        row2 = int(eval(input('Row number [stop] :')))
        for i in range(row1, row2 + 1):
            del order[row1]
    if mode1 == 8:
        if type(tempsave) == type(1):
            if tempsave != -1:
                row1 = int(eval(input('Paste before number:')))
                row2 = int(eval(input('How many times?:')))
                for i in range(row2):
                    order.insert(row1, tempsave)
            else:
                print('nothing in temp, copy first')
                eval(input('Continue'))
        else:
            row1 = int(eval(input('Paste before number:')))
            row2 = int(eval(input('How many times?:')))
            for i in range(row2):
                for i in reversed(list(range(len(tempsave)))):
                    order.insert(row1, tempsave[i])
    if mode1 == 9:
        row1 = int(eval(input('Swap line: ')))
        row2 = int(eval(input('with line: ')))
        ordertemp = order[row1]
        order[row1] = order[row2]
        order[row2] = ordertemp
    if mode1 == 10:
        nspintemp = int(eval(input('Double potential (1=no,2=yes) ')))
        natomtemp = int(open('scoef').readlines()[0])
        filedata = open('scoef').readlines()[1:natomtemp + 1]
        listnew = []
        if mode == 'shape':
            shapeorder = iatom2shape_2016()
            print(shapeorder)
        for line in filedata:
            if (len(line.split()) > 1):
                if mode == 'pot':
                    listnew.append(int(line.split()[3]) - 1)
                    if (nspintemp == 2):
                        listnew.append(int(line.split()[3]) - 1)
                elif mode == 'shape':
                    listnew.append(shapeorder[int(line.split()[3]) - 1] - 1)
        print('new order list')
        print(listnew)
        print(len(listnew))
        order = listnew
    if mode1 == 11:
        npot = len(order)
        if npot % 2 != 0:
            print('number of potentials is odd')
            exit()
        print(list(range(0, len(order), 2)))
        for i in range(0, len(order), 2):
            ordertemp = order[i]
            order[i] = order[i + 1]
            order[i + 1] = ordertemp
    if mode1 == 12:
        listnew = eval(input('New list :').split())
        makeorder = lambda x: int(x) - 1
        order = list(map(makeorder, listnew))
    if mode1 == 13:
        ordernew = []
        for item in order:
            ordernew.append(item)
            ordernew.append(item)
        order = ordernew
    if mode1 == 14:
        filedata = open('scoef').readlines()[1:]
        listnew = []
        if mode == 'shape':
            shapeorder = iatom2shape_2016()
            print(shapeorder)
        for line in filedata:
            if (len(line.split()) > 1):
                if mode == 'pot':
                    listnew.append((int(line.split()[3]) - 1) * 2)
                    listnew.append((int(line.split()[3]) - 1) * 2 + 1)
                elif mode == 'shape':
                    listnew.append(shapeorder[int(line.split()[3]) - 1] - 1)
        print('new order list')
        print(listnew)
        print(len(listnew))
        order = listnew
    if mode1 == 15:
        print('mode 15 chosen')
        if len(order) % 2 != 0:
            exit('ERROR: odd number of potentials')
        for i in range(len(order) // 2):
            print(i)
            head1, pot1 = read_pot_values(index1[2 * i + 0], index2[2 * i + 0])
            head2, pot2 = read_pot_values(index1[2 * i + 1], index2[2 * i + 1])
            if head1[1:7] != head2[1:7]:
                print(head1, head2)
                exit('ERROR: potential header for spin up and down do not match!!!')
            vc, vs = calc_coulomb_spin_pot(pot1, pot2)
            head1[0] = head1[0].replace('up', 'averaged')
            head1[0] = head1[0].replace('UP', 'averaged')
            head1[0] = head1[0].replace('down', 'averaged')
            head1[0] = head1[0].replace('DOWN', 'averaged')
            headnew1 = combine_header(head1, head2, 0)
            headnew2 = combine_header(head1, head2, 0)
            print(len(headnew1), len(head1), len(head2))
            newdat1 = pot2newdata(headnew1, vc)
            newdat2 = pot2newdata(headnew2, vc)
            print(len(headnew1), len(head1), len(head2), len(newdat1), len(newdat2), len(vc), len(vs))
            print(i, index1, index2)
            data[index1[2 * i]:index2[2 * i]] = newdat1
            data[index1[2 * i + 1]:index2[2 * i + 1]] = newdat2
    if mode1 == 16:
        if len(order) % 2 != 0:
            exit('ERROR: odd number of potentials')
        print('scale magnetic moment up: newpot_1/2 = (v_1+v_2)/2 +/- alpha*(v_1-v_2)/2')
        alpha = float(eval(input('alpha: ')))
        for i in range(len(order) // 2):
            print(i)
            head1, pot1 = read_pot_values(index1[2 * i + 0], index2[2 * i + 0])
            head2, pot2 = read_pot_values(index1[2 * i + 1], index2[2 * i + 1])
            if head1[1:7] != head2[1:7]:
                print(head1, head2)
                exit('ERROR: potential header for spin up and down do not match!!!')
            vc, vs = calc_coulomb_spin_pot(pot1, pot2)
            pot_new1 = combine_potentials(vc, vs, alpha)
            pot_new2 = combine_potentials(vc, vs, -alpha)
            headnew1 = combine_header(head1, head2, alpha)
            headnew2 = combine_header(head2, head1, alpha)
            print(len(headnew1), len(head1), len(head2))
            newdat1 = pot2newdata(headnew1, pot_new1)
            newdat2 = pot2newdata(headnew2, pot_new2)
            print(len(headnew1), len(head1), len(head2), len(newdat1), len(newdat2), len(vc), len(vs))
            print(i, index1, index2)
            data[index1[2 * i]:index2[2 * i]] = newdat1
            data[index1[2 * i + 1]:index2[2 * i + 1]] = newdat2

datanew = []
for i in range(len(order)):
    for ii in range(index1[order[i]], index2[order[i]] + 1):
        datanew.append(data[ii])
if mode == 'pot':
    with open('potential_new', 'w') as f:
        f.writelines(datanew)
elif mode == 'shape':
    # add header to shapefun_new
    tmp = datanew
    datanew = []
    datanew.append('   %i\n' % (len(order)))
    datanew.append('  1.000000000000E+00\n')
    datanew += tmp
    with open('shapefun_new', 'w') as f:
        f.writelines(datanew)
else:
    print('error')
    exit()



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains the dictionary with all defined tasks for the outxml_parser.
The entries in the TASK_DEFINITION dict specify how to parse specific attributes tags.

This needs to be maintained if the specifications do not work for a new schema version
because of changed attribute names for example.

Each entry in the TASK_DEFINITION dict can contain a series of keys, which by default
correspond to the keys in the output dictionary

The following keys are expected in each entry:
    :param parse_type: str, defines which methods to use when extracting the information
    :param path_spec: dict with all the arguments that should be passed to get_tag_xpath
                      or get_attrib_xpath to get the correct path
    :param subdict: str, if present the parsed values are put into this key in the output dictionary
    :param overwrite_last: bool, if True no list is inserted and each entry overwrites the last

For the allAttribs parse_type there are more keys that can appear:
    :param base_value: str, optional. If given the attribute
                       with this name will be inserted into the key from the task_definition
                       all other keys are formatted as {task_key}_{attribute_name}
    :param ignore: list of str, these attributes will be ignored
    :param overwrite: list of str, these attributes will not create a list and overwrite any value
                      that might be there
    :param flat: bool, if False the dict parsed from the tag is inserted as a dict into the correspondin key
                       if True the values will be extracted and put into the output dictionary with the
                       format {task_key}_{attribute_name}

Each task entry can have additional keys to specify, when to perform the task.
These are denoted with underscores in their names and are all optional:

    :param _general: bool, default False. If True the parsing is not performed for each iteration on the
                     iteration node but beforehand and on the root node
    :param _modes: list of tuples, sets conditions for the keys in fleur_modes to perform the task
                   .e.g. [('jspins', 2), ('soc', True)] means only perform this task for a magnetic soc calculation
    :param _minimal: bool, default False, denotes task to perform when minimal_mode=True is passed to the parser
    :param _special: bool, default False, If true these tasks are not added by default and need to be added manually
    :param _conversions: list of str, gives the names of functions in fleur_outxml_conversions to perform after parsing


The following keys are special at the moment:
    - ```fleur_modes``` specifies how to identify the type of the calculation (e.g. SOC, magnetic, lda+u)
      this is used to determine, whether additional things should be parsed

Following is the current specification of tasks

.. literalinclude:: ../../../masci_tools/io/parsers/fleur/default_parse_tasks.py
   :language: python
   :lines: 66-
   :linenos:

"""

__working_out_versions__ = {'0.34'}
__base_version__ = '0.34'

TASKS_DEFINITION = {
    #--------Definitions for general info from outfile (start, endtime, number_iterations)--------
    'general_out_info': {
        '_general': True,
        '_minimal': True,
        '_conversions': ['calculate_walltime'],
        'creator_name': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'version',
                'not_contains': 'git'
            }
        },
        'creator_target_architecture': {
            'parse_type': 'text',
            'path_spec': {
                'name': 'targetComputerArchitectures'
            }
        },
        'output_file_version': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'fleurOutputVersion'
            }
        },
        'number_of_iterations': {
            'parse_type': 'numberNodes',
            'path_spec': {
                'name': 'iteration'
            }
        },
        'number_of_atoms': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'nat'
            }
        },
        'number_of_atom_types': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'ntype'
            }
        },
        'number_of_kpoints': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'count',
                'contains': 'numericalParameters'
            }
        },
        'start_date': {
            'parse_type': 'allAttribs',
            'path_spec': {
                'name': 'startDateAndTime'
            },
            'ignore': ['zone'],
            'flat': False,
        },
        'end_date': {
            'parse_type': 'allAttribs',
            'path_spec': {
                'name': 'endDateAndTime'
            },
            'ignore': ['zone'],
            'flat': False,
        }
    },
    #--------Defintions for general info from input section of outfile (kmax, symmetries, ..)--------
    'general_inp_info': {
        '_general': True,
        '_minimal': True,
        'title': {
            'parse_type': 'text',
            'path_spec': {
                'name': 'comment'
            }
        },
        'kmax': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'Kmax'
            }
        },
        'gmax': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'Gmax'
            }
        },
        'number_of_spin_components': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'jspins'
            }
        },
        'number_of_symmetries': {
            'parse_type': 'numberNodes',
            'path_spec': {
                'name': 'symOp'
            }
        },
        'number_of_species': {
            'parse_type': 'numberNodes',
            'path_spec': {
                'name': 'species'
            }
        },
        'film': {
            'parse_type': 'exists',
            'path_spec': {
                'name': 'filmPos'
            }
        },
    },
    #--------Defintions for lda+u info from input section (species, ldau tags)--------
    'ldau_info': {
        '_general': True,
        '_modes': [('ldau', True)],
        '_conversions': ['convert_ldau_definitions'],
        'parsed_ldau': {
            'parse_type': 'allAttribs',
            'path_spec': {
                'name': 'ldaU',
                'contains': 'species'
            },
            'subdict': 'ldau_info',
            'flat': False,
            'only_required': True
        },
        'ldau_species': {
            'parse_type': 'parentAttribs',
            'path_spec': {
                'name': 'ldaU',
                'contains': 'species'
            },
            'subdict': 'ldau_info',
            'flat': False,
            'only_required': True
        }
    },
    #--------Defintions for relaxation info from input section (bravais matrix, atompos)
    #--------for Bulk and film
    'bulk_relax_info': {
        '_general': True,
        '_modes': [('relax', True), ('film', False)],
        '_conversions': ['convert_relax_info'],
        'lat_row1': {
            'parse_type': 'text',
            'path_spec': {
                'name': 'row-1',
                'contains': 'bulkLattice/bravais'
            }
        },
        'lat_row2': {
            'parse_type': 'text',
            'path_spec': {
                'name': 'row-2',
                'contains': 'bulkLattice/bravais'
            }
        },
        'lat_row3': {
            'parse_type': 'text',
            'path_spec': {
                'name': 'row-3',
                'contains': 'bulkLattice/bravais'
            }
        },
        'atom_positions': {
            'parse_type': 'text',
            'path_spec': {
                'name': 'relPos'
            }
        },
        'position_species': {
            'parse_type': 'parentAttribs',
            'path_spec': {
                'name': 'relPos'
            },
            'flat': False,
            'only_required': True
        },
        'element_species': {
            'parse_type': 'allAttribs',
            'path_spec': {
                'name': 'species'
            },
            'flat': False,
            'ignore': ['vcaAddCharge', 'magField']
        },
    },
    'film_relax_info': {
        '_general': True,
        '_modes': [('relax', True), ('film', True)],
        '_conversions': ['convert_relax_info'],
        'lat_row1': {
            'parse_type': 'text',
            'path_spec': {
                'name': 'row-1',
                'contains': 'filmLattice/bravais'
            }
        },
        'lat_row2': {
            'parse_type': 'text',
            'path_spec': {
                'name': 'row-2',
                'contains': 'filmLattice/bravais'
            }
        },
        'lat_row3': {
            'parse_type': 'text',
            'path_spec': {
                'name': 'row-3',
                'contains': 'filmLattice/bravais'
            }
        },
        'atom_positions': {
            'parse_type': 'text',
            'path_spec': {
                'name': 'filmPos'
            }
        },
        'position_species': {
            'parse_type': 'parentAttribs',
            'path_spec': {
                'name': 'filmPos'
            },
            'flat': False,
            'only_required': True
        },
        'element_species': {
            'parse_type': 'allAttribs',
            'path_spec': {
                'name': 'species'
            },
            'flat': False,
            'ignore': ['vcaAddCharge', 'magField']
        },
    },
    #----General iteration tasks
    # iteration number
    # total energy (only total or also contributions, also lda+u correction)
    # distances (nonmagnetic and magnetic, lda+u density matrix)
    # charges (total, interstitial, mt sphere)
    # fermi energy and bandgap
    # magnetic moments
    # orbital magnetic moments
    # forces
    'iteration_number': {
        '_minimal': True,
        'number_of_iterations_total': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'overallNumber'
            },
            'overwrite_last': True,
        }
    },
    'total_energy': {
        '_minimal': True,
        '_conversions': ['convert_total_energy'],
        'energy_hartree': {
            'parse_type': 'singleValue',
            'path_spec': {
                'name': 'totalEnergy'
            }
        },
    },
    'distances': {
        '_minimal': True,
        'density_convergence': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'distance',
                'tag_name': 'chargeDensity'
            }
        },
        'density_convergence_units': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'units',
                'tag_name': 'densityConvergence',
            },
            'overwrite_last': True,
        }
    },
    'magnetic_distances': {
        '_minimal': True,
        '_modes': [('jspin', 2)],
        'overall_density_convergence': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'distance',
                'tag_name': 'overallChargeDensity'
            }
        },
        'spin_density_convergence': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'distance',
                'tag_name': 'spinDensity'
            }
        }
    },
    'total_energy_contributions': {
        'sum_of_eigenvalues': {
            'parse_type': 'singleValue',
            'path_spec': {
                'name': 'sumOfEigenvalues'
            },
            'only_required': True
        },
        'energy_core_electrons': {
            'parse_type': 'singleValue',
            'path_spec': {
                'name': 'coreElectrons',
                'contains': 'sumOfEigenvalues'
            },
            'only_required': True
        },
        'energy_valence_electrons': {
            'parse_type': 'singleValue',
            'path_spec': {
                'name': 'valenceElectrons'
            },
            'only_required': True
        },
        'charge_den_xc_den_integral': {
            'parse_type': 'singleValue',
            'path_spec': {
                'name': 'chargeDenXCDenIntegral'
            },
            'only_required': True
        },
    },
    'ldau_energy_correction': {
        '_modes': [('ldau', True)],
        'ldau_energy_correction': {
            'parse_type': 'singleValue',
            'path_spec': {
                'name': 'dftUCorrection'
            },
            'subdict': 'ldau_info',
            'only_required': True
        },
    },
    'nmmp_distances': {
        '_minimal': True,
        '_modes': [('ldau', True)],
        'density_matrix_distance': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'distance',
                'contains': 'ldaUDensityMatrixConvergence'
            },
            'subdict': 'ldau_info'
        },
    },
    'fermi_energy': {
        'fermi_energy': {
            'parse_type': 'singleValue',
            'path_spec': {
                'name': 'FermiEnergy'
            },
        }
    },
    'bandgap': {
        '_modes': [('bz_integration', 'hist')],
        'bandgap': {
            'parse_type': 'singleValue',
            'path_spec': {
                'name': 'bandgap'
            },
        }
    },
    'magnetic_moments': {
        '_modes': [('jspin', 2)],
        'magnetic_moments': {
            'parse_type': 'allAttribs',
            'path_spec': {
                'name': 'magneticMoment'
            },
            'base_value': 'moment',
            'ignore': ['atomType']
        }
    },
    'orbital_magnetic_moments': {
        '_modes': [('jspin', 2), ('soc', True)],
        'orbital_magnetic_moments': {
            'parse_type': 'allAttribs',
            'path_spec': {
                'name': 'orbMagMoment'
            },
            'base_value': 'moment',
            'ignore': ['atomType']
        }
    },
    'forces': {
        '_minimal': True,
        '_modes': [('relax', True)],
        '_conversions': ['convert_forces'],
        'force_units': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'units',
                'tag_name': 'totalForcesOnRepresentativeAtoms'
            },
            'overwrite_last': True
        },
        'parsed_forces': {
            'parse_type': 'allAttribs',
            'path_spec': {
                'name': 'forceTotal'
            },
            'flat': False,
            'only_required': True
        }
    },
    'charges': {
        '_conversions': ['calculate_total_magnetic_moment'],
        'spin_dependent_charge': {
            'parse_type': 'allAttribs',
            'path_spec': {
                'name': 'spinDependentCharge',
                'contains': 'allElectronCharges',
                'not_contains': 'fixed'
            },
            'only_required': True
        },
        'total_charge': {
            'parse_type': 'singleValue',
            'path_spec': {
                'name': 'totalCharge',
                'contains': 'allElectronCharges',
                'not_contains': 'fixed'
            },
            'only_required': True
        }
    },
    #-------Tasks for forcetheorem Calculations
    # DMI, JIJ, MAE, SSDISP
    'forcetheorem_dmi': {
        '_special': True,
        'dmi_force': {
            'parse_type': 'allAttribs',
            'path_spec': {
                'name': 'Entry',
                'contains': 'DMI'
            }
        },
        'dmi_force_qs': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'qpoints',
                'contains': 'Forcetheorem_DMI'
            }
        },
        'dmi_force_angles': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'Angles',
                'contains': 'Forcetheorem_DMI'
            }
        },
        'dmi_force_units': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'units',
                'contains': 'Forcetheorem_DMI'
            }
        }
    },
    'forcetheorem_ssdisp': {
        '_special': True,
        'spst_force': {
            'parse_type': 'allAttribs',
            'path_spec': {
                'name': 'Entry',
                'contains': 'SSDISP'
            }
        },
        'spst_force_qs': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'qvectors',
                'contains': 'Forcetheorem_SSDISP'
            }
        },
        'spst_force_units': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'units',
                'contains': 'Forcetheorem_SSDISP'
            }
        }
    },
    'forcetheorem_mae': {
        '_special': True,
        'mae_force': {
            'parse_type': 'allAttribs',
            'path_spec': {
                'name': 'Angle',
                'contains': 'MAE'
            }
        },
        'mae_force_units': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'units',
                'contains': 'Forcetheorem_MAE'
            }
        }
    },
    'forcetheorem_jij': {
        '_special': True,
        'jij_force': {
            'parse_type': 'allAttribs',
            'path_spec': {
                'name': 'Config',
                'contains': 'JIJ'
            }
        },
        'jij_force_units': {
            'parse_type': 'attrib',
            'path_spec': {
                'name': 'units',
                'contains': 'Forcetheorem_JIJ'
            }
        }
    }
}
